import asyncio
import logging
import platform

import psutil

from prtg_pyprobe.sensors.helpers import (
    SensorProbeData,
    SensorDefinition,
)
from prtg_pyprobe.sensors.sensor import SensorPSUtilBase


class Sensor(SensorPSUtilBase):
    @property
    def name(self) -> str:
        return "Temperature (Probe)"

    @property
    def kind(self) -> str:
        return "mpprobetemperature"

    @property
    def definition(self) -> dict:
        probe_health_def = SensorDefinition(
            name=self.name,
            kind=self.kind,
            description="Sensor used to monitor the temperature of the Python probe.",
            sensor_help="This sensor monitors the temp of the probe itself. "
            "This is always the case even if added to a different device than the Probe Device.",
            tag="mpprobetemperaturesensor",
        )

        probe_health_def.add_group(self.temperature_settings_group)
        return probe_health_def.data

    async def work(self, task_data: dict, q: asyncio.Queue) -> None:
        probe_temp_data = SensorProbeData(sensor_id=task_data["sensorid"])

        try:
            uname = platform.uname()
            probe_temp_data.message = f"Temperatures on {uname.system}, everything OK"

            temperatures = self.get_system_temperatures(fahrenheit=True if task_data["celfar"] == 2 else False)
            if temperatures:
                for name, entries in temperatures.items():
                    for entry in entries:
                        probe_temp_data.add_temperature_channel(entry.label or name, entry.current)

            else:
                probe_temp_data.error = "Exception"
                probe_temp_data.error_code = 1
                probe_temp_data.message = (
                    "The sensor could not find temperature sensors on your probe machine." " See log for details"
                )
                logging.exception("No temperature sensors could be found by PSUtil")

        except psutil.Error:
            probe_temp_data.error = "Exception"
            probe_temp_data.error_code = 1
            probe_temp_data.message = "Probe temperature check failed. See log for details"
            logging.exception("Probe temperature could not be determined")

        await q.put(probe_temp_data.data)
        q.task_done()



import asyncio
import logging
import platform

import psutil

from prtg_pyprobe.sensors.helpers import SensorProbeData, SensorDefinition, SensorDefinitionGroup
from prtg_pyprobe.sensors.sensor import SensorPSUtilBase


class Sensor(SensorPSUtilBase):
    @property
    def name(self) -> str:
        return "Memory (Probe)"

    @property
    def kind(self) -> str:
        return "mpprobememory"

    @property
    def definition(self) -> dict:
        probe_health_def = SensorDefinition(
            name=self.name,
            kind=self.kind,
            description="Sensor used to monitor the memory usage of the system on which the python probe is running.",
            sensor_help="Sensor used to monitor the memory usage of the system on which the python probe is running."
            "This is always the case even if added to a different device than the Probe Device.",
            tag="mpprobememorysensor",
        )
        memory_group = SensorDefinitionGroup(name="probe_memory", caption="Probe Memory")
        probe_health_def.add_group(memory_group)
        return probe_health_def.data

    async def work(self, task_data: dict, q: asyncio.Queue) -> None:
        probe_health_data = SensorProbeData(sensor_id=task_data["sensorid"])
        try:
            uname = platform.uname()
            probe_health_data.message = f"Memory usage on {uname.system}, everything OK"
            vmemory, swapmemory = self.get_memory_usage()

            probe_health_data.add_channel(name="Memory Total", mode="integer", kind="BytesMemory", value=vmemory.total)
            probe_health_data.add_channel(
                name="Memory Available",
                mode="integer",
                kind="BytesMemory",
                value=vmemory.available,
            )
            probe_health_data.add_channel(
                name="Swap Total",
                mode="integer",
                kind="BytesMemory",
                value=swapmemory.total,
            )
            probe_health_data.add_channel(name="Swap Free", mode="integer", kind="BytesMemory", value=swapmemory.free)

        except psutil.Error:
            probe_health_data.error = "Exception"
            probe_health_data.error_code = 1
            probe_health_data.message = "Probe memory check failed. See log for details"
            logging.exception("Probe memory information could not be determined")
        await q.put(probe_health_data.data)
        q.task_done()



import logging
from abc import ABC
from collections import namedtuple
from typing import Union

import psutil
from pysnmp.error import PySnmpError
from pysnmp.hlapi import asyncio as pysnmp_asyncio
from pysnmp.hlapi.asyncio import (
    CommunityData,
    SnmpEngine,
    ContextData,
    UdpTransportTarget,
    ObjectType,
    ObjectIdentity,
)

from prtg_pyprobe.sensors.helpers import SensorDefinitionGroup


class SensorBase(ABC):
    """Abstract Base Class, from which every sensor should inherit."""

    def __getitem__(self, kind):
        return self.kind

    @property
    def name(self) -> str:
        """Abstract property name of SensorBase.

        Raises:
            NotImplementedError: Has to be overwritten in implementation
        """
        raise NotImplementedError

    @property
    def kind(self) -> str:
        """Abstract property kind of SensorBase.

        Raises:
            NotImplementedError: Has to be overwritten in implementation
        """
        raise NotImplementedError

    @property
    def definition(self) -> dict:
        """Abstract property definition of SensorBase.

        Raises:
            NotImplementedError: Has to be overwritten in implementation
        """
        raise NotImplementedError

    async def work(self, *args, **kwargs) -> None:
        """Abstract method work of SensorBase.

        Args:
            *args:
            **kwargs:

        Raises:
            NotImplementedError: Has to be overwritten in implementation

        """
        raise NotImplementedError


class SensorPSUtilBase(SensorBase, ABC):
    """Base class for all sensors utilizing `psutil`."""

    @property
    def temperature_settings_group(self):
        """Add Group and Field for temperature settings"""
        probe_health_def_group_temperature = SensorDefinitionGroup(name="temperature", caption="Temperature Settings")
        probe_health_def_group_temperature.add_field(
            field_type="radio",
            name="celfar",
            caption="Choose between Celsius or Fahrenheit display",
            help="Choose wether you want to return the value in Celsius or Fahrenheit",
            options={"1": "Celsius", "2": "Fahrenheit"},
            default="1",
        )
        return probe_health_def_group_temperature

    @property
    def diskspace_settings_group(self):
        """Add Group and Field for diskspace settings"""
        probe_health_def_group_diskspace = SensorDefinitionGroup(name="diskspace", caption="Disk Space Settings")
        probe_health_def_group_diskspace.add_field(
            field_type="radio",
            name="diskfull",
            caption="Full Display of all disk space values",
            help="Choose wether you want to get all disk space data or only percentages.",
            options={"1": "Percentages", "2": "Full Information"},
            default="1",
        )
        return probe_health_def_group_diskspace

    @staticmethod
    def get_system_temperatures(fahrenheit: bool) -> Union[dict, None]:
        """Get system temperature.

        Args:
            fahrenheit (bool): Use Fahrenheut or Celcius as unit.

        Returns:
            Union: Return either a dict or None.
        """
        if not hasattr(psutil, "sensors_temperatures"):
            return None
        temperatures = psutil.sensors_temperatures(fahrenheit=fahrenheit)
        if temperatures == {}:
            return None

        return temperatures

    @staticmethod
    def get_system_partitions() -> list:
        """Get partitions.

        Returns:
            list: A list of partitions.
        """
        return psutil.disk_partitions()

    @staticmethod
    def get_partition_usage(partition: list) -> namedtuple:
        """Get disk usage of partitions.

        Args:
            partition (list): A list of partitions.

        Returns:
            namedtuple: A named tuple containing disk usage data.

        """
        return psutil.disk_usage(partition[1])

    @staticmethod
    def get_memory_usage() -> tuple:
        """Get memory usage.

        Returns:
            tuple: A tuple containing virtual memory and swap usage.

        """
        vmemory = psutil.virtual_memory()
        swapmemory = psutil.swap_memory()

        return vmemory, swapmemory

    @staticmethod
    def get_cpu_load() -> tuple:
        """Get CPU load average.

        Returns:
            tuple: A tuple containing the load average.
        """
        return psutil.getloadavg()


class SensorSNMPBase(SensorBase, ABC):
    """Base class for all SNMP sensors."""

    @property
    def snmp_specific_settings_group(self):
        """Add Group and Field for SNMP related settings."""
        snmp_group_snmp_specifications = SensorDefinitionGroup(name="snmpspecific", caption="SNMP Specific")
        snmp_group_snmp_specifications.add_field(
            field_type="edit",
            name="community",
            caption="Community String",
            required="1",
            help="Please enter the community string.",
        )
        snmp_group_snmp_specifications.add_field(
            field_type="integer",
            name="port",
            caption="SNMP Port",
            required="1",
            help="Please enter SNMP Port.",
        )
        snmp_group_snmp_specifications.add_field(
            field_type="radio",
            name="snmp_version",
            caption="SNMP Version",
            required="1",
            help="Choose your SNMP Version",
            options={"0": "V1", "1": "V2c", "2": "V3"},
            default="1",
        )
        return snmp_group_snmp_specifications

    async def get(
        self,
        target: str,
        oids: list,
        credentials: CommunityData,
        port: int = 161,
        engine: SnmpEngine = SnmpEngine(),
        context: ContextData = ContextData(),
    ) -> list:
        """SNMP single get.

        Args:
            target (str): The target of the SNMP request.
            oids (list): A list of OIDs to query.
            credentials (CommunityData): The PySnmp authentication data.
            port (int): The port the target listens for SNMP requests.
            engine (SnmpEngine): PySnmp SNMP Engine.
            context (ContextData): PySnmp context data

        Returns:
            list: A list of OIDs and values.

        Raises:
            PySnmpError: Raised when something goes wrong.
        """
        (error_indication, error_status, error_index, var_binds,) = await pysnmp_asyncio.getCmd(
            engine,
            credentials,
            UdpTransportTarget((target, port)),
            context,
            *self.construct_object_types(oids),
        )
        if error_indication:
            logging.error(error_indication)
            raise PySnmpError(error_indication)
        elif error_status:
            msg = f"{error_status.prettyPrint()} at {error_index and var_binds[int(error_index) - 1][0] or '?'}"
            logging.error(msg)
            raise PySnmpError(msg)
        return var_binds

    async def get_bulk(
        self,
        target: str,
        oids: list,
        credentials: CommunityData,
        port: int = 161,
        engine: SnmpEngine = SnmpEngine(),
        context: ContextData = ContextData(),
    ):
        """SNMP bulk get (getnext).

        Args:
            target (str): The target of the SNMP request.
            oids (list): A list of OIDs to query.
            credentials (CommunityData): The PySnmp authentication data.
            port (int): The port the target listens for SNMP requests.
            engine (SnmpEngine): PySnmp SNMP Engine.
            context (ContextData): PySnmp context data

        Returns:
            list: A list of OIDs and values.

        Raises:
            PySnmpError: Raised when something goes wrong.
        """
        # todo:  Probably we won't need bulk get
        result = []
        var_binds = self.construct_object_types(oids)
        initial_var_binds = var_binds
        while True:
            (error_indication, error_status, error_index, var_bind_table,) = await pysnmp_asyncio.nextCmd(
                engine,
                credentials,
                UdpTransportTarget((target, port)),
                context,
                *var_binds,
                lookupValues=True,
                lookupNames=True,
            )

            if error_indication:
                logging.error(error_indication)
                raise PySnmpError(error_indication)
            elif error_status:
                msg = f"{error_status.prettyPrint()} at {error_index and var_binds[int(error_index) - 1][0] or '?'}"
                logging.error(msg)
                raise PySnmpError(msg)
            else:
                for var_bind_row in var_bind_table:
                    result.append(var_bind_row)

                var_binds = var_bind_table[-1]
                if pysnmp_asyncio.isEndOfMib(var_binds):
                    break

                for idx, var_bind in enumerate(var_binds):
                    if initial_var_binds[0][idx].isPrefixOf(var_bind[0]):
                        break

                else:
                    break

        return result

    @staticmethod
    def construct_object_types(list_of_oids: list) -> list:
        """Construnc objects from given OIDs

        Args:
            list_of_oids: A list of OIDs.

        Returns:
            list: A list of PySnmp Object Identities.

        """
        object_types = []
        for oid in list_of_oids:
            object_types.append(ObjectType(ObjectIdentity(oid)))
        return object_types

    @staticmethod
    def cast(value: str) -> Union[str, float, int]:
        """A brute force cast method.

        Args:
            value (str): A string.

        Returns:
            Union: A hopefully cast value.
        """
        try:
            return int(value)
        except (ValueError, TypeError):
            try:
                return float(value)
            except (ValueError, TypeError):
                return str(value)



import asyncio
import itertools
import logging

from pysnmp.hlapi.asyncio import SnmpEngine

from prtg_pyprobe.communication.communication_http_api import PRTGHTTPApi


async def run_monitoring_tasks(
    sensors_avail: list,
    taskdata: list,
    prtg_api: PRTGHTTPApi,
    snmp_engine: SnmpEngine,
    async_queue: asyncio.Queue,
    config: dict,
):
    """Matches tasks to sensors and does the acutal monitoring.

    Args:
        sensors_avail (list): List of available sensors.
        taskdata (list): List of tasks from the PRTG Core.
        prtg_api (PRTGHTTPApi): A PRTG HTTP API object.
        snmp_engine (SnmpEngine): SNMP Engine object.
        async_queue (asyncio.Queue): An AsyncIO Queue.
        config (dict): The probe config as dict.

    """
    monitoring_task_list = []
    task_chunk_size = int(config["probe_task_chunk_size"])
    task_chunks = [taskdata[x : x + task_chunk_size] for x in range(0, len(taskdata), task_chunk_size)]
    for chunk in task_chunks:
        for sensor, task in itertools.product(sensors_avail, chunk):
            if task["kind"] == sensor.kind:
                if "mpsnmp" in task["kind"]:
                    monitoring_task_list.append(
                        asyncio.create_task(sensor.work(task_data=task, q=async_queue, snmp_engine=snmp_engine))
                    )
                else:
                    monitoring_task_list.append(asyncio.create_task(sensor.work(task_data=task, q=async_queue)))
        logging.info(f"Current Queue Size: {async_queue.qsize()}")

        logging.debug(f"Running Tasks in Event Loop (before join): {asyncio.all_tasks()}")
        await asyncio.gather(*monitoring_task_list)
        results = []
        while not async_queue.qsize() == 0:
            results.append(await async_queue.get())
        [mt.cancel() for mt in monitoring_task_list]
        logging.debug(f"Running Tasks in Event Loop (after join): {asyncio.all_tasks()}")
        chunk_response = await prtg_api.send_data(sensor_response_data=results)
        if chunk_response.status_code == 200:
            logging.info("Monitoring data successfully sent to PRTG Core.")


async def monitoring(
    tasks_list: list, sensor_objects: list, prtg_api: PRTGHTTPApi, snmp_engine: SnmpEngine, config: dict
):
    """
    An additional wrapper around ``run_monitoring_tasks``
    Args:
        tasks_list (list):
        sensor_objects (list):
        prtg_api (PRTGHTTPApi): A PRTG HTTP API object.
        snmp_engine (SnmpEngine): SNMP Engine object.
        config (dict):

    Returns:
        list: A list of monitoring results

    """
    logging.info("Initializing Queue")
    async_queue = asyncio.Queue()

    res = await run_monitoring_tasks(
        sensors_avail=sensor_objects,
        taskdata=tasks_list,
        snmp_engine=snmp_engine,
        async_queue=async_queue,
        prtg_api=prtg_api,
        config=config,
    )
    logging.debug(f"Running Tasks in Event Loop (after results): {asyncio.all_tasks()}")
    return res



import datetime
import hashlib
import os
import subprocess
import sys
from distutils.sysconfig import get_python_lib

import PyInquirer
import click
import daemon.daemon as ppd
import psutil

from prtg_pyprobe.run import main as pyprobe_main
from prtg_pyprobe.utils import config
from prtg_pyprobe.utils.config import configuration_questions
from prtg_pyprobe.utils.defaults import LOG_PATH, CONFIG_PATH


def hash_access_key(key: str) -> str:
    """Hash the access key needed to connect to the PRTG Core.

    Args:
        key (string):  Alphanumeric string.

    Returns:
        string: SHA-1 hashed version of the string.

    """
    key = key.encode("utf-8")
    return hashlib.sha1(key).hexdigest()


def check_systemd():
    """Check if a system uses systemd as startup system.

    Returns:
        bool: True id systemd. False otherwise.

    """
    result = subprocess.run(["ps", "--no-headers", "-o", "comm", "1"], capture_output=True)
    if result.stdout == b"systemd\n":
        return True
    return False


@click.group()
def pyprobe():
    pass


@click.command(name="start")
def dstart():
    """Start a probe daemon."""
    site_packages = get_python_lib()
    work_dir = f"{site_packages}/prtg_pyprobe/"
    click.echo("pyprobe daemon starting")
    with ppd.DaemonContext(
        working_directory=work_dir,
    ):
        pyprobe_main()


@click.command(name="stop")
def dstop():
    """Stop the probe daemon."""
    running_procs = psutil.process_iter()
    for process in running_procs:
        if any("bin/pyprobe" in string for string in process.cmdline()):
            process.terminate()
    click.echo("pyprobe daemon terminated")


@click.command(name="status")
def dstatus():
    """Get status of probe daemon."""
    running_procs = psutil.process_iter()
    for process in running_procs:
        if any("bin/pyprobe" in string for string in process.cmdline()):
            if not any("status" in string for string in process.cmdline()):
                run_since = datetime.datetime.fromtimestamp(process.create_time()).strftime("%Y-%m-%d %H:%M:%S")
                click.echo(f"--- Status {process.name()} ---")
                click.echo(f"PID: {process.pid}")
                click.echo(f"Running as: {process.username()}")
                click.echo(f"Running since: {run_since}")


@click.command()
def configure():
    """Creates or overwrites the configuration.

    Raises:
        KeyboardInterrupt: Keyboard Interrupt is raised to cancel the configuration.

    """
    try:
        if not os.getenv("DEV"):
            if not os.path.exists(LOG_PATH):
                os.mkdir(path=LOG_PATH, mode=666)
            if not os.path.exists(CONFIG_PATH):
                os.mkdir(path=CONFIG_PATH, mode=644)
        config_name = "config.yml"
        if os.path.isfile(f"{CONFIG_PATH}{config_name}"):
            click.echo("\033[93m Completing this config wizard will overwrite your current config! \033[93m")
            cont = PyInquirer.prompt(
                {
                    "type": "confirm",
                    "message": "Do you want to continue?",
                    "name": "continue",
                    "default": True,
                },
            )
            if not cont["continue"]:
                raise KeyboardInterrupt
        answers = PyInquirer.prompt(configuration_questions)
    except KeyboardInterrupt:
        print("Exiting configuration wizard")
        sys.exit(0)
    if answers != {}:
        answers["probe_protocol_version"] = "1"
        answers["probe_access_key_hashed"] = hash_access_key(answers["probe_access_key"])
        probe_cfg = config.ProbeConfig(path=f"{CONFIG_PATH}{config_name}")
        probe_cfg.write(answers)
        click.echo(f"Config written to {CONFIG_PATH}{config_name}")


@click.group()
def daemon():
    pass


@click.group()
def service():
    pass


@click.command(name="install")
def sinstall():
    """Installs the pyprobe service via systemd."""
    if check_systemd():
        site_packages = get_python_lib()
        with open(f"{site_packages}/prtg_pyprobe/utils/pyprobe.service", "r") as fs:
            pyprobe_service = fs.read()
        with open("/etc/systemd/system/pyprobe.service", "w") as fsc:
            pyprobe_service = pyprobe_service.format(
                sys.executable,
                f"{site_packages}/prtg_pyprobe/run.py",
                f"{site_packages}/prtg_pyprobe",
            )
            fsc.write(pyprobe_service)
        subprocess.run(["systemctl", "daemon-reload"])
        subprocess.run(["systemctl", "enable", "pyprobe.service"])
        click.echo("Service install done")
    else:
        click.echo("Service only supported on systems running systemd")


@click.command(name="start")
def sstart():
    """Starts the pyprobe service."""
    if check_systemd():
        subprocess.run(["service", "pyprobe", "start"])
        click.echo("Service started")
    else:
        click.echo("Service only supported on systems running systemd")


@click.command(name="stop")
def sstop():
    """Stops the pyprobe service."""
    if check_systemd():
        subprocess.run(["service", "pyprobe", "stop"])
        click.echo("Service stopped")
    else:
        click.echo("Service only supported on systems running systemd")


@click.command(name="status")
def sstatus():
    """Status of the pyprobe service."""
    if check_systemd():
        subprocess.run(["service", "pyprobe", "status"])
    else:
        click.echo("Service only supported on systems running systemd")


daemon.add_command(dstart)
daemon.add_command(dstop)
daemon.add_command(dstatus)

service.add_command(sinstall)
service.add_command(sstart)
service.add_command(sstop)
service.add_command(sstatus)


pyprobe.add_command(configure)
pyprobe.add_command(service)
pyprobe.add_command(daemon)



import asyncio
import json
import logging
from json import JSONDecodeError
from typing import Any, Union

import httpx
import dpath.util

from prtg_pyprobe.sensors.helpers import (
    SensorData,
    SensorDefinitionGroup,
    SensorDefinition,
)
from prtg_pyprobe.sensors.sensor import SensorBase


class Sensor(SensorBase):
    @property
    def name(self) -> str:
        return "HTTP REST"

    @property
    def kind(self) -> str:
        return "mphttprest"

    @property
    def definition(self) -> dict:
        http_def = SensorDefinition(
            name=self.name,
            kind=self.kind,
            description="Monitors a REST response",
            sensor_help="Monitors the availability and REST response of an HTTP target.",
            tag="mphttprestsensor",
        )
        http_def_group_target = SensorDefinitionGroup(name="targetinfo", caption="Target Information")
        http_def_group_rest = SensorDefinitionGroup(name="restinfo", caption="REST Information")
        http_def_group_target.add_field_timeout(10, 10, 900)
        http_def_group_target.add_field(
            field_type="edit",
            name="target",
            caption="REST Target",
            required="1",
            default="https://",
            help="Enter a valid REST target to monitor.  The format should be http(s)://xyz.com",
        )
        http_def_group_target.add_field(
            field_type="radio",
            name="request_type",
            caption="Request Type",
            required="1",
            options={"GET": "GET", "POST": "POST"},
            default="GET",
        )
        http_def_group_target.add_field(
            field_type="edit",
            name="headers",
            caption="HTTP Headers",
            default="",
            help="Enter HTTP Headers as dictionary. e.g. {'my header':'my value'}",
        )
        http_def_group_target.add_field(
            field_type="edit",
            name="post_data",
            caption="POST Data",
            default="",
            help="Enter POST data as dictionary. e.g. {'my header':'my value'}. Only used when Request Type is POST",
        )
        http_def_group_target.add_field(
            field_type="edit",
            name="acceptable_status_codes",
            caption="Acceptable HTTP Status Codes",
            default="200",
            help="Enter acceptable HTTP Status codes. Format is comma separated. Leave empty for no filter.",
        )
        http_def_group_rest.add_field(
            field_type="edit",
            name="dpathselector",
            caption="Selector",
            default="",
            help="Enter a valid selector for a value in XPATH notation. "
            "Leave empty if you only want to measure the API response time.",
        )
        http_def_group_rest.add_field(
            field_type="radio",
            name="value_type",
            caption="Request Type",
            required="1",
            options={"float": "Float", "integer": "Integer"},
            default="float",
        )
        http_def.add_group(group=http_def_group_target)
        http_def.add_group(group=http_def_group_rest)
        return http_def.data

    async def work(self, task_data: dict, q: asyncio.Queue) -> None:
        http_rest_data = SensorData(sensor_id=task_data["sensorid"])
        try:
            headers = None
            post_data = None
            acceptable_status_codes = []
            if task_data["acceptable_status_codes"]:
                acceptable_status_codes = [int(i) for i in task_data["acceptable_status_codes"].split(",")]
            if task_data["headers"]:
                try:
                    headers = json.loads(task_data["headers"])
                except JSONDecodeError:
                    raise
            if task_data["post_data"] and task_data["request_type"] == "POST":
                try:
                    post_data = json.loads(task_data["post_data"])
                except JSONDecodeError:
                    raise

            async with httpx.AsyncClient() as client:
                request = client.build_request(
                    method=task_data["request_type"],
                    url=task_data["target"],
                    headers=headers,
                    data=post_data,
                )
                response = await client.send(request=request, timeout=int(task_data["timeout"]))

            if len(acceptable_status_codes) and response.status_code not in acceptable_status_codes:
                raise httpx.HTTPStatusError("Status Code not in Filter list", request=request, response=response)

            response_dict = response.json()

            if not task_data["dpathselector"] == "":
                response_value = dpath.util.get(response_dict, task_data["dpathselector"])
                http_rest_data.add_channel(
                    name="REST return value",
                    mode=task_data["value_type"],
                    kind="Custom",
                    customunit="",
                    value=self.cast(response_value, task_data["value_type"]),
                )

            http_rest_data.add_channel(
                name="Response Time",
                mode="float",
                kind="TimeResponse",
                value=float(response.elapsed.microseconds / 1000),
            )
            http_rest_data.add_channel(
                name="HTTP Status Code",
                mode="integer",
                kind="Custom",
                customunit="",
                value=response.status_code,
            )
            http_rest_data.message = (
                f"HTTP target returned {response.status_code}, Request Type: {task_data['request_type']}"
            )

        except TypeError:
            http_rest_data.error = "Exception"
            http_rest_data.error_code = 1
            http_rest_data.message = "Response not in correct format."
            logging.exception("Response not in correct format")
        except JSONDecodeError:
            http_rest_data.error = "Exception"
            http_rest_data.error_code = 1
            http_rest_data.message = "POST Data, HTTP Headers or response not in correct format."
            logging.exception("POST data, headers or response not in correct format")
        except KeyError:
            http_rest_data.error = "Exception"
            http_rest_data.error_code = 1
            http_rest_data.message = "Element not found in response."
            logging.exception("Element not found in response. Xpath selector wrong?")
        except ValueError:
            http_rest_data.error = "Exception"
            http_rest_data.error_code = 1
            http_rest_data.message = "Query returned more than one element."
            logging.exception("Too much elements found in response. Xpath selector wrong?")
        except httpx.HTTPStatusError:
            http_rest_data.error = "Exception"
            http_rest_data.error_code = 1
            http_rest_data.message = f"Status Code {response.status_code} not in acceptable list"
            logging.exception("HTTP Status did not match any in the acceptable list.")
        except httpx.HTTPError:
            http_rest_data.error = "Exception"
            http_rest_data.error_code = 1
            http_rest_data.message = "HTTP request failed. See log for details"
            logging.exception("There was an error in httpx.")
        except CastException:
            http_rest_data.error = "Exception"
            http_rest_data.error_code = 1
            http_rest_data.message = "Could not cast the value xpath returned to the correct format."
            logging.exception("Casting error.")

        await q.put(http_rest_data.data)
        q.task_done()

    @staticmethod
    def cast(value: Any, target: str) -> Union[int, float]:
        try:
            value_cast = value
            if target == "integer":
                value_cast = int(value)
            elif target == "float":
                value_cast = float(value)
            return value_cast
        except ValueError:
            raise CastException(value=value, target=target)


class CastException(Exception):
    def __init__(self, value, target):
        self._message = f"Could not convert {value} to {target}"
        super().__init__(self._message)



import pathlib

import pytest
from psutil._common import shwtemp

from prtg_pyprobe.communication.communication_http_api import PRTGHTTPApi
from prtg_pyprobe.sensors import (
    sensor_ping,
    sensor_port,
    sensor_http,
    sensor_http_rest,
    sensor_probe_memory,
    sensor_probe_health,
    sensor_snmp_custom,
    sensor_snmp_cpuload,
    sensor_snmp_traffic,
    sensor_traffic,
    sensor_dns,
    sensor_dns_reverse,
    sensor_probe_temperature,
    sensor_probe_diskspace,
    sensor_external_ip,
    sensor_cpuload,
    sensor_s3_bucket_total,
    sensor_prometheus_exporter,
)
from prtg_pyprobe.sensors.helpers import (
    SensorDefinition,
    SensorDefinitionGroup,
    SensorData,
)
from prtg_pyprobe.utils.config import (
    IPDNSValidator,
    PortRangeValidator,
    NotEmptyValidator,
    GUIDValidator,
    BaseIntervalValidator,
    BaseChunkSizeValidator,
    DirectoryWriteableValidator,
    ProbeConfig,
)
from prtg_pyprobe.utils.sensor_loader import load_sensor_modules, create_sensor_objects


@pytest.fixture()
def list_sensor_modules():
    print(pathlib.Path().absolute())
    return load_sensor_modules(sensor_path="../prtg_pyprobe/sensors")


@pytest.fixture()
def list_sensor_objects():
    print(pathlib.Path().absolute())
    return create_sensor_objects(sensor_path="../prtg_pyprobe/sensors")


@pytest.fixture()
def ip_dns_validator():
    return IPDNSValidator()


@pytest.fixture()
def port_range_validator():
    return PortRangeValidator()


@pytest.fixture()
def not_empty_validator():
    return NotEmptyValidator()


@pytest.fixture()
def guid_validator():
    return GUIDValidator()


@pytest.fixture()
def base_interval_validator():
    return BaseIntervalValidator()


@pytest.fixture()
def base_chunk_size_validator():
    return BaseChunkSizeValidator()


@pytest.fixture()
def directory_writeable_validator():
    return DirectoryWriteableValidator()


@pytest.fixture()
def probe_config():
    return ProbeConfig(path="config.yml")


@pytest.fixture()
def probe_config_dict():
    probe_cfg_dict = {
        "disable_ssl_verification": True,
        "log_file_location": "./pyprobe.log",
        "log_level": "INFO",
        "probe_access_key": "miniprobe",
        "probe_access_key_hashed": "cd7b773e2ce4205e9f5907b157f3d26495c5b373",
        "probe_base_interval": "60",
        "probe_task_chunk_size": "20",
        "probe_gid": "1BFB9273-08D7-43AF-B535-18E4A767BA34",
        "probe_name": "Python Mini Probe",
        "probe_protocol_version": "1",
        "prtg_server_ip_dns": "test.prtg.com",
        "prtg_server_port": "443",
    }
    return probe_cfg_dict


@pytest.fixture()
def prtg_api(probe_config_dict):
    return PRTGHTTPApi(probe_config=probe_config_dict, backoff_factor=0.1)


@pytest.fixture()
def sensor_definition():
    definition = SensorDefinition(
        name="Testsensor",
        kind="testkind",
        description="This is a Test",
        sensor_help="Test Help",
        tag="testsensor",
        default="1",
    )
    return definition


@pytest.fixture()
def sensor_definition_group():
    definition_group = SensorDefinitionGroup(name="testgroup", caption="TestCaption")
    return definition_group


@pytest.fixture()
def sensor_data():
    sensor_data = SensorData(sensor_id="1234")
    return sensor_data


@pytest.fixture()
def ping_sensor():
    return sensor_ping.Sensor()


@pytest.fixture()
def port_sensor():
    return sensor_port.Sensor()


@pytest.fixture()
def http_sensor():
    return sensor_http.Sensor()


@pytest.fixture()
def http_rest_sensor():
    return sensor_http_rest.Sensor()


@pytest.fixture()
def prometheus_exporter_sensor():
    return sensor_prometheus_exporter.Sensor()


@pytest.fixture()
def probehealth_sensor():
    return sensor_probe_health.Sensor()


@pytest.fixture()
def probe_memory_sensor():
    return sensor_probe_memory.Sensor()


@pytest.fixture()
def probetemp_sensor():
    return sensor_probe_temperature.Sensor()


@pytest.fixture()
def probe_disk_space_sensor():
    return sensor_probe_diskspace.Sensor()


@pytest.fixture()
def snmp_custom_sensor():
    return sensor_snmp_custom.Sensor()


@pytest.fixture()
def snmp_traffic_sensor():
    return sensor_snmp_traffic.Sensor()


@pytest.fixture()
def snmp_cpu_load_sensor():
    return sensor_snmp_cpuload.Sensor()


@pytest.fixture()
def traffic_sensor():
    return sensor_traffic.Sensor()


@pytest.fixture()
def dns_sensor():
    return sensor_dns.Sensor()


@pytest.fixture()
def dns_reverse_sensor():
    return sensor_dns_reverse.Sensor()


@pytest.fixture()
def external_ip_sensor():
    return sensor_external_ip.Sensor()


@pytest.fixture()
def cpu_load_sensor():
    return sensor_cpuload.Sensor()


@pytest.fixture()
def s3_total_sensor():
    return sensor_s3_bucket_total.Sensor()


@pytest.fixture()
def snmp_custom_sensor_taskdata():
    return {
        "sensorid": "1234",
        "community": "public",
        "port": "161",
        "snmp_version": "0",
        "host": "127.0.0.1",
        "custom_oid": "1.3.6.1.1.2.3.4",
        "custom_unit": "test",
        "value_type": "integer",
    }


@pytest.fixture()
def snmp_traffic_sensor_taskdata():
    return {
        "sensorid": "1234",
        "community": "public",
        "port": "161",
        "snmp_version": "0",
        "host": "127.0.0.1",
        "ifindex": "1",
    }


@pytest.fixture()
def snmp_cpuload_sensor_taskdata():
    return {
        "sensorid": "1234",
        "community": "public",
        "port": "161",
        "snmp_version": "0",
        "host": "127.0.0.1",
    }


@pytest.fixture()
def task_data_ping():
    return [
        {
            "sensorid": "1234",
            "host": "paessler.com",
            "timeout": "5",
            "kind": "mpping",
            "target_port": "443",
            "ping_count": "4",
            "ping_type": "tcp",
        }
    ]


@pytest.fixture()
def task_data_snmp():
    return [
        {
            "sensorid": "1234",
            "host": "127.0.0.1",
            "community": "public",
            "port": "161",
            "kind": "mpsnmpcustom",
            "snmp_version": "1",
            "custom_oid": "1.3.6.1",
            "custom_unit": "#",
            "value_type": "integer",
        }
    ]


@pytest.fixture()
def sensor_exception_message():
    return {
        "sensorid": "1234",
        "message": "Overwrite me in test case",
        "error": "Exception",
        "code": 1,
    }


async def asyncio_open_connection(*args, **kwargs):
    reader = ReturnObjectAsyncioOpenConnection()
    writer = ReturnObjectAsyncioOpenConnection()
    return reader, writer


class ReturnObjectAsyncioOpenConnection(object):
    def close(self):
        return

    async def wait_closed(self):
        return


@pytest.fixture()
def psutil_system_temperatures():
    return {
        "acpitz": [shwtemp(label="", current=60.5, high=84.0, critical=84.0)],
        "coretemp": [
            shwtemp(label="Package id 0", current=53.0, high=100.0, critical=100.0),
            shwtemp(label="Core 0", current=53.0, high=100.0, critical=100.0),
            shwtemp(label="Core 1", current=52.0, high=100.0, critical=100.0),
            shwtemp(label="Package id 0", current=53.0, high=100.0, critical=100.0),
            shwtemp(label="Core 0", current=53.0, high=100.0, critical=100.0),
            shwtemp(label="Core 1", current=52.0, high=100.0, critical=100.0),
        ],
        "dell_smm": [
            shwtemp(label="CPU", current=60.0, high=None, critical=None),
            shwtemp(label="Ambient", current=39.0, high=None, critical=None),
            shwtemp(label="Ambient", current=32.0, high=None, critical=None),
            shwtemp(label="GPU", current=16.0, high=None, critical=None),
            shwtemp(label="Other", current=43.0, high=None, critical=None),
        ],
    }


class DnsAnswer(object):
    @property
    def address(self):
        return "1.2.3.4"

    @property
    def target(self):
        return "mytarget"

    @property
    def preference(self):
        return "mypref"

    @property
    def exchange(self):
        return "myex"

    @property
    def mname(self):
        return "mname"

    @property
    def rname(self):
        return "rname"

    @property
    def serial(self):
        return "serial"

    @property
    def refresh(self):
        return 60

    @property
    def expire(self):
        return 60

    @property
    def strings(self):
        return [b"string1", b"string2"]


async def dns_answer(*args, **kwargs):
    return [DnsAnswer()]



import asyncio
import json
import logging
from json import JSONDecodeError
from sys import getsizeof

import httpx

from prtg_pyprobe.sensors.helpers import (
    SensorData,
    SensorDefinitionGroup,
    SensorDefinition,
)
from prtg_pyprobe.sensors.sensor import SensorBase


class Sensor(SensorBase):
    @property
    def name(self) -> str:
        return "HTTP"

    @property
    def kind(self) -> str:
        return "mphttp"

    @property
    def definition(self) -> dict:
        http_def = SensorDefinition(
            name=self.name,
            kind=self.kind,
            description="Monitors an HTTP target",
            sensor_help="Monitors the availability and response time of an HTTP target",
            tag="mphttpsensor",
        )
        http_def_group_target = SensorDefinitionGroup(name="targetinfo", caption="Target Information")
        http_def_group_target.add_field_timeout(10, 10, 900)
        http_def_group_target.add_field(
            field_type="edit",
            name="target",
            caption="HTTP Target",
            required="1",
            default="https://",
            help="Enter a valid HTTP target to monitor.  The format should be http(s)://xyz.com",
        )
        http_def_group_target.add_field(
            field_type="radio",
            name="request_type",
            caption="Request Type",
            required="1",
            options={"GET": "GET", "POST": "POST", "HEAD": "HEAD"},
            default="GET",
        )
        http_def_group_target.add_field(
            field_type="edit",
            name="headers",
            caption="HTTP Headers",
            default="",
            help="Enter HTTP Headers as dictionary. e.g. {'my header':'my value'}",
        ),
        http_def_group_target.add_field(
            field_type="edit",
            name="post_data",
            caption="POST Data",
            default="",
            help="Enter POST data as dictionary. e.g. {'my header':'my value'}. Only used when Request Type is POST",
        )
        http_def_group_target.add_field(
            field_type="edit",
            name="acceptable_status_codes",
            caption="Acceptable HTTP Status Codes",
            default="200",
            help="Enter acceptable HTTP Status codes. Format is comma separated. Leave empty for no filter.",
        )
        http_def.add_group(group=http_def_group_target)
        return http_def.data

    async def work(self, task_data: dict, q: asyncio.Queue) -> None:
        http_data = SensorData(sensor_id=task_data["sensorid"])
        # todo: basic auth
        try:
            headers = None
            post_data = None
            acceptable_status_codes = []
            if task_data["acceptable_status_codes"]:
                acceptable_status_codes = [int(i) for i in task_data["acceptable_status_codes"].split(",")]
            if task_data["headers"]:
                try:
                    headers = json.loads(task_data["headers"])
                except JSONDecodeError:
                    raise
            if task_data["post_data"] and task_data["request_type"] == "POST":
                try:
                    post_data = json.loads(task_data["post_data"])
                except JSONDecodeError:
                    raise

            async with httpx.AsyncClient() as client:
                request = client.build_request(
                    method=task_data["request_type"],
                    url=task_data["target"],
                    headers=headers,
                    data=post_data,
                )
                response = await client.send(request=request, timeout=int(task_data["timeout"]))

            if len(acceptable_status_codes) and response.status_code not in acceptable_status_codes:
                raise httpx.HTTPStatusError("Status Code not in Filter list", request=request, response=response)

            http_data.add_channel(
                name="Response Time",
                mode="float",
                kind="TimeResponse",
                value=float(response.elapsed.microseconds / 1000),
            )
            http_data.add_channel(
                name="HTTP Status Code",
                mode="integer",
                kind="Custom",
                customunit="",
                value=response.status_code,
            )
            http_data.add_channel(
                name="Response Size",
                mode="float",
                kind="BytesFile",
                value=float(getsizeof(response.text)),
            )
            http_data.message = (
                f"HTTP target returned {response.status_code}, Request Type: {task_data['request_type']}"
            )

        except httpx.HTTPStatusError:
            http_data.error = "Exception"
            http_data.error_code = 1
            http_data.message = f"Status Code {response.status_code} not in acceptable list"
            logging.exception("HTTP Status did not match any in the acceptable list.")
        except httpx.HTTPError:
            http_data.error = "Exception"
            http_data.error_code = 1
            http_data.message = "HTTP request failed. See log for details"
            logging.exception("There was an error in httpx.")
        except JSONDecodeError:
            http_data.error = "Exception"
            http_data.error_code = 1
            http_data.message = "POST Data or HTTP Headers not in correct format."
            logging.exception("POST data or headers not in correct format")

        await q.put(http_data.data)
        q.task_done()



import asyncio
import ctypes
import logging
import os
import socket
import time
from statistics import mean, pstdev

import aioping

from prtg_pyprobe.sensors.helpers import (
    SensorData,
    SensorDefinitionGroup,
    SensorDefinition,
)
from prtg_pyprobe.sensors.sensor import SensorBase


class Sensor(SensorBase):
    @property
    def name(self) -> str:
        return "Ping (TCP/ICMP)"

    @property
    def kind(self) -> str:
        return "mpping"

    @property
    def definition(self) -> dict:
        ping_def = SensorDefinition(
            name=self.name,
            kind=self.kind,
            description="Monitors the availability of a target using TCP or ICMP",
            sensor_help="Monitors the availability of a target using TCP or ICMP",
            tag="mppingsensor",
        )
        ping_def_group_settings = SensorDefinitionGroup(name="pingsettings", caption="Ping Settings")
        ping_def_group_settings.add_field_timeout(5, 1, 300)
        ping_def_group_settings.add_field(
            field_type="integer",
            name="ping_count",
            caption="Ping Count",
            required="1",
            default=4,
            minimum=1,
            maximum=20,
            help="Enter the count of Ping requests PRTG will send to the device during an interval",
        )
        ping_def_group_settings.add_field(
            field_type="integer",
            name="target_port",
            caption="Port (TCP mode only)",
            required="1",
            default=443,
            minimum=1,
            maximum=65534,
            help="This is the port that you want to monitor if it's open or closed. Only used for TCP Ping",
        )
        ping_def_group_settings.add_field(
            field_type="radio",
            name="ping_type",
            caption="Ping Type",
            required="1",
            help="Choose your Ping type. Note that for ICMP the pyprobe must run as root!",
            options={"tcp": "TCP", "icmp": "ICMP"},
            default="tcp",
        )
        ping_def.add_group(group=ping_def_group_settings)
        return ping_def.data

    async def work(self, task_data: dict, q: asyncio.Queue) -> None:
        ping_data = SensorData(sensor_id=task_data["sensorid"])
        try:
            ping_raw_data = []
            for i in range(int(task_data["ping_count"])):
                await asyncio.sleep(0.1)
                if task_data["ping_type"] == "tcp":
                    ping_future = asyncio.open_connection(host=task_data["host"], port=task_data["target_port"])
                    start = time.time()
                    reader, writer = await asyncio.wait_for(ping_future, int(task_data["timeout"]))
                    writer.close()
                    await writer.wait_closed()
                    end = time.time()
                    time_ms = (end - start) * 1000
                    ping_raw_data.append(round(time_ms, 4))
                elif task_data["ping_type"] == "icmp":
                    if not self._is_admin():
                        raise PermissionError("The pyprobe must run as root/administrator to use ICMP Ping")
                    single_ping = await aioping.ping(
                        dest_addr=task_data["host"],
                        timeout=int(task_data["timeout"]),
                        family=socket.AddressFamily.AF_INET,
                    )
                    ping_raw_data.append(single_ping * 1000)

            ping_data.message = f"OK. Ping Mode {task_data['ping_type']}"
            ping_data.add_channel(
                name="Ping Time Min",
                mode="float",
                kind="TimeResponse",
                value=float(min(ping_raw_data)),
            )
            ping_data.add_channel(
                name="Ping Time Avg",
                mode="float",
                kind="TimeResponse",
                value=float(mean(ping_raw_data)),
            )
            ping_data.add_channel(
                name="Ping Time Max",
                mode="float",
                kind="TimeResponse",
                value=float(max(ping_raw_data)),
            )
            ping_data.add_channel(
                name="Ping Time MDEV",
                mode="float",
                kind="TimeResponse",
                value=float(pstdev(ping_raw_data)),
            )
        except ConnectionError:
            ping_data.error = "Exception"
            ping_data.error_code = 1
            ping_data.message = "Ping check failed. See log for details"
            logging.exception("There has been a connection error.")
        except asyncio.TimeoutError:
            ping_data.error = "Exception"
            ping_data.error_code = 1
            ping_data.message = "Ping Timeout. See log for details"
            logging.exception("There has been a timeout.")
        except TimeoutError:
            ping_data.error = "Exception"
            ping_data.error_code = 1
            ping_data.message = "Ping Timeout. See log for details"
            logging.exception("There has been a timeout.")
        except socket.gaierror:
            ping_data.error = "Exception"
            ping_data.error_code = 1
            ping_data.message = "Ping check failed. See log for details"
            logging.exception("There has been a connection error.")
        except PermissionError:
            ping_data.error = "Exception"
            ping_data.error_code = 1
            ping_data.message = "The pyprobe must run as root/administrator to use ICMP Ping"
            logging.exception("ICMP Ping uses RAW sockets. You must be root to use them.")
        except OSError:
            ping_data.error = "Exception"
            ping_data.error_code = 1
            ping_data.message = "Target device/port probably not available."
            logging.exception(
                "Something on the network level went wrong. " "Target host not reachable on selected port."
            )
        await q.put(ping_data.data)
        q.task_done()

    @staticmethod
    def _is_admin():
        try:
            is_admin = os.getuid() == 0
        except AttributeError:
            is_admin = ctypes.windll.shell32.IsUserAnAdmin()
        return is_admin



import time

import aioboto3
import logging
from asyncio import Queue

from botocore.exceptions import ClientError

from prtg_pyprobe.sensors.helpers import SensorDefinition, SensorDefinitionGroup, SensorData
from prtg_pyprobe.sensors.sensor import SensorBase


class Sensor(SensorBase):
    @property
    def name(self) -> str:
        return "S3 Bucket Total"

    @property
    def kind(self) -> str:
        return "mps3buckettotal"

    @property
    def definition(self) -> dict:
        s3_def = SensorDefinition(
            name=self.name,
            kind=self.kind,
            description="Sensor for monitoring the total amount of your S3 buckets.",
            sensor_help="This returns the total amount of buckets in your AWS account.",
            tag="mps3buckettotal",
        )
        s3_def_group = SensorDefinitionGroup(name="s3_bucket_total", caption="S3 Bucket Sensor Settings")
        s3_def_group.add_field(field_type="edit", name="aws_access_key", caption="AWS Access Key ID")
        s3_def_group.add_field(field_type="password", name="aws_secret_key", caption="AWS Secret Key")
        s3_def.add_group(s3_def_group)
        return s3_def.data

    async def work(self, task_data: dict, q: Queue) -> None:
        s3_bucket_data = SensorData(sensor_id=task_data["sensorid"])

        try:
            start = time.time()
            async with aioboto3.resource(
                "s3", aws_access_key_id=task_data["aws_access_key"], aws_secret_access_key=task_data["aws_secret_key"]
            ) as s3:
                all_buckets = s3.buckets.all()
                i = 0
                async for _ in all_buckets:
                    i += 1

                s3_bucket_data.add_channel(
                    name="Total Buckets", mode="integer", kind="Custom", customunit="buckets", value=i
                )

                s3_bucket_data.message = f"Your AWS account has {i} buckets."

                end = (time.time() - start) * 1000
                s3_bucket_data.add_channel(name="Total Query Time", mode="float", kind="TimeResponse", value=end)

        except ClientError:
            s3_bucket_data.error = "Exception"
            s3_bucket_data.error_code = 1
            s3_bucket_data.message = "S3 bucket sensor failed. See log for details"
            logging.exception("The S3 bucket sensor couldn't contact AWS properly")

        await q.put(s3_bucket_data.data)
        q.task_done()



import importlib
import pkgutil

from prtg_pyprobe.sensors.sensor import SensorBase


def load_sensor_modules(sensor_path: str = "sensors") -> dict:
    """Basic plugin loader with enforcing of naming conventions (modules have to start with "sensor_").

    Args:
        sensor_path (:obj:`str` optional): Path where sensor modules are located.

    Returns:
        dict: Dict of sensor modules.

    """
    # plugin discovery with enforcing of naming convention
    usable_plugins = {}
    discovered_plugins = {
        name: importlib.import_module(f"prtg_pyprobe.sensors.{name}")
        for finder, name, ispkg in pkgutil.iter_modules([sensor_path])
        if name.startswith("sensor_")
    }

    # check if the Plugin is inheriting from a common base class
    for name, mod in discovered_plugins.items():
        if issubclass(mod.Sensor, SensorBase):
            usable_plugins[name] = mod
    return discovered_plugins


def create_sensor_objects(*args, **kwargs) -> list:
    """Create sensor objects.

    Args:
        *args: Variable length argument list. Usually empty.
        **kwargs: Arbitrary keyword arguments. Usually empty.

    Returns:
        list: List of sensor objects.

    """
    # creating objects from loaded modules
    obj_list = []
    for name, module in load_sensor_modules(*args, **kwargs).items():
        obj_list.append(module.Sensor())
    return obj_list



import functools
from typing import Any

from cerberus import Validator

unit_kind_allowed = [
    "BytesBandwidth",
    "BytesMemory",
    "BytesDisk",
    "BytesFile",
    "TimeResponse",
    "TimeSeconds",
    "TimeHours",
    "Temperature",
    "Percent",
    "Count",
    "CPU",
    "Custom",
]
field_type_allowed = ["edit", "integer", "password", "radio"]
log_level_allowed = ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]

validation_schemes = {
    "SensorDefinitionGroupField": {
        "name": {"type": "string", "required": True, "empty": False},
        "type": {
            "type": "string",
            "required": True,
            "allowed": field_type_allowed,
            "empty": False,
        },
        "caption": {"type": "string", "required": True, "empty": False},
        "required": {"type": "string", "allowed": ["yes", "1", "true"]},
        "help": {"type": "string"},
        "default": {"type": ["string", "integer"]},
        "minimum": {"type": "integer"},
        "maximum": {"type": "integer"},
        "options": {
            "type": "dict",
            "keysrules": {"type": "string"},
            "valuesrules": {"type": "string"},
        },
    },
    "SensorDefinitionGroup": {
        "name": {
            "type": "string",
            "required": True,
            "empty": False,
            "regex": "^[a-z0-9_-]{5,40}",
        },
        "caption": {"type": "string", "required": True, "empty": False},
        "fields": {"type": "list", "required": False, "empty": True},
    },
    "SensorDefinition": {
        "name": {"type": "string", "required": True, "empty": False},
        "kind": {"type": "string", "required": True, "empty": False},
        "description": {"type": "string", "required": True, "empty": False},
        "help": {"type": "string", "required": True, "empty": False},
        "tag": {"type": "string", "required": True, "empty": False},
        "default": {"type": "string", "allowed": ["yes", "1", "true"]},
        "groups": {"type": "list", "required": True, "empty": True},
    },
    "SensorData": {
        "sensorid": {"type": "string", "required": True, "empty": False},
        "message": {"type": "string"},
        "channel": {"type": "list"},
        "error": {"type": "string", "dependencies": ["message", "code"]},
        "code": {"type": "integer"},
    },
    "SensorProbeData": {
        "sensorid": {"type": "string", "required": True, "empty": False},
        "message": {"type": "string"},
        "channel": {"type": "list"},
        "error": {"type": "string", "dependencies": ["message", "code"]},
        "code": {"type": "integer"},
    },
    "SensorDataChannel": {
        "name": {"type": "string", "required": True, "empty": False},
        "mode": {
            "type": "string",
            "required": True,
            "allowed": ["integer", "float", "counter"],
            "empty": False,
        },
        "value": {"type": ["integer", "float"]},
        "unit": {"type": "string", "allowed": unit_kind_allowed},
        "kind": {"type": "string", "allowed": unit_kind_allowed},
        "customunit": {"type": "string"},
    },
    "ProbeConfig": {
        "log_file_location": {"type": "string", "empty": True},
        "log_level": {"type": "string", "allowed": log_level_allowed, "empty": False},
        "disable_ssl_verification": {"type": "boolean", "empty": False},
        "probe_access_key": {"type": "string", "empty": False},
        "probe_access_key_hashed": {"type": "string", "empty": False},
        "probe_base_interval": {"type": "string", "empty": False},
        "probe_task_chunk_size": {"type": "string", "empty": False},
        "probe_gid": {
            "type": "string",
            "empty": False,
            "regex": "([a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}){1}",
        },
        "probe_name": {"type": "string", "empty": False},
        "probe_protocol_version": {"type": "string", "empty": False, "allowed": ["1"]},
        "prtg_server_ip_dns": {"type": "string", "empty": False},
        "prtg_server_port": {"type": "string", "empty": False},
    },
}


class ValidationFailedError(Exception):
    """Raised when validation against scheme failed."""

    def __init__(self, validator):
        """Constructor of ValidationFailedError.

        Args:
            validator (cerberus.validator): Cerberus validator.

        """
        self._validator = validator
        self._message = self._validator.errors
        super().__init__(self._message)


class ValidationNotPossibleError(Exception):
    """Raised when validation is not possible."""

    def __init__(self):
        super().__init__("Cannot validate, no validation schema defined.")


def validate(func: Any):
    """Decorator to validate dictionaries.

    Args:
        func (Any): Decorated function.

    Returns:
        Any: Wrapped function.

    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        value = func(*args, **kwargs)
        try:
            schema = validation_schemes[type(args[0]).__name__]
        except KeyError:
            raise ValidationNotPossibleError
        return validation(value, schema)

    return wrapper


def validation(val: dict, scheme: dict) -> dict:
    """Validation of a dict against a cerberus scheme.

    Args:
        val (dict): Input dictionary.
        scheme (dict): Cerberus validation scheme.

    Returns:
        dict: Input dictionary if validation succeeded.

    Raises:
        ValidationFailedError: Raised if the validation fails.
    """
    v = Validator(scheme)
    if v.validate(val):
        return val
    else:
        raise ValidationFailedError(v)



import asyncio
import logging
import platform

import psutil

from prtg_pyprobe.sensors.helpers import (
    SensorProbeData,
    SensorDefinition,
)
from prtg_pyprobe.sensors.sensor import SensorPSUtilBase


class Sensor(SensorPSUtilBase):
    @property
    def name(self) -> str:
        return "Disk Usage (Probe)"

    @property
    def kind(self) -> str:
        return "mpprobediskusage"

    @property
    def definition(self) -> dict:
        probe_health_def = SensorDefinition(
            name=self.name,
            kind=self.kind,
            description="Sensor used to monitor the disk space usage of the Python probe.",
            sensor_help="This sensor monitors the disk space usage of the probe itself."
            "This is always the case even if added to a different device than the Probe Device.",
            tag="mpprobediskusage",
        )

        probe_health_def.add_group(self.diskspace_settings_group)
        return probe_health_def.data

    async def work(self, task_data: dict, q: asyncio.Queue) -> None:
        probe_disk_data = SensorProbeData(sensor_id=task_data["sensorid"])

        try:
            uname = platform.uname()
            disk_partitions = self.get_system_partitions()
            for partition in disk_partitions:
                part_disk_usage = self.get_partition_usage(partition)
                probe_disk_data.add_disk_space_percentage_use(partition, part_disk_usage)

                if task_data["diskfull"] == "2":
                    probe_disk_data.add_disk_space_details_channel("Total", partition, part_disk_usage[0])
                    probe_disk_data.add_disk_space_details_channel("Used", partition, part_disk_usage[1])
                    probe_disk_data.add_disk_space_details_channel("Free", partition, part_disk_usage[2])

            probe_disk_data.message = f"Disk usage on {uname.system}, everything OK"

        except psutil.Error:
            probe_disk_data.error = "Exception"
            probe_disk_data.error_code = 1
            probe_disk_data.message = "Probe disk space check failed. See log for details"
            logging.exception("Probe disk space could not be determined")

        await q.put(probe_disk_data.data)
        q.task_done()



import asyncio
import logging
import time

import dns.asyncresolver
import dns.reversename
from dns.exception import DNSException

from prtg_pyprobe.sensors.helpers import (
    SensorData,
    SensorDefinitionGroup,
    SensorDefinition,
)
from prtg_pyprobe.sensors.sensor import SensorBase


class Sensor(SensorBase):
    @property
    def name(self) -> str:
        return "DNS Reverse"

    @property
    def kind(self) -> str:
        return "mpdnsptr"

    @property
    def definition(self) -> dict:
        dns_re_def = SensorDefinition(
            name=self.name,
            kind=self.kind,
            description="DNS PTR Lookup",
            sensor_help="The DNS sensor monitors a Domain Name Service (DNS) server. "
            "It resolves a domain name and compares it to a given IP address.",
            tag="mpdnssensor",
        )
        dns_re_def_group_dns_spec = SensorDefinitionGroup(name="dns_specific", caption="DNS Specific")
        dns_re_def_group_dns_spec.add_field_timeout(5, 1, 900)
        dns_re_def_group_dns_spec.add_field(
            field_type="integer",
            name="port",
            caption="Port",
            required="1",
            default=53,
            minimum=1,
            maximum=65535,
            help="Enter the port on which the DNS service of the parent device is running.",
        )
        dns_re_def_group_dns_spec.add_field(
            field_type="edit",
            name="ip",
            caption="IP Address",
            required="1",
            help="Enter a DNS name to resolve.",
        )
        dns_re_def.add_group(dns_re_def_group_dns_spec)
        return dns_re_def.data

    async def work(self, task_data: dict, q: asyncio.Queue):
        dns_re_data = SensorData(sensor_id=task_data["sensorid"])
        try:
            start = time.time()
            resolver = dns.asyncresolver.Resolver()
            reverse_addr = dns.reversename.from_address(task_data["ip"])
            resolver.nameservers = [f"{task_data['host']}"]
            resolver.port = int(task_data["port"])
            res = await resolver.resolve(reverse_addr, rdtype="PTR")
            msg = f"Query Type: PTR for {task_data['ip']}. Result: "
            end = (time.time() - start) * 1000
            for rdata in res:
                msg += f"{rdata.target}, "
            dns_re_data.message = msg
            dns_re_data.add_channel(
                name="Repsonse Time",
                mode="float",
                kind="TimeResponse",
                value=float(end),
            )
        except DNSException:
            dns_re_data.error = "Exception"
            dns_re_data.error_code = 1
            dns_re_data.message = "DNS check failed. See log for details"
            logging.exception("Error in dnspython.")
        await q.put(dns_re_data.data)
        q.task_done()



import asyncio
import logging
import socket
import time

from prtg_pyprobe.sensors.helpers import (
    SensorData,
    SensorDefinitionGroup,
    SensorDefinition,
)
from prtg_pyprobe.sensors.sensor import SensorBase


class Sensor(SensorBase):
    @property
    def name(self) -> str:
        return "Port"

    @property
    def kind(self) -> str:
        return "mpport"

    @property
    def definition(self) -> dict:
        port_def = SensorDefinition(
            name=self.name,
            kind=self.kind,
            description="Monitors the availability of a port on a target system",
            sensor_help="Monitors the availability of a port on a target system",
            tag="mpportsensor",
        )
        port_def_group_port_spec = SensorDefinitionGroup(name="portspecific", caption="Port specific")

        port_def_group_port_spec.add_field_timeout(60, 1, 900)
        port_def_group_port_spec.add_field(
            field_type="integer",
            name="targetport",
            caption="Port",
            required="1",
            default=110,
            minimum=1,
            maximum=65534,
            help="This is the port that you want to monitor if it's open or closed",
        )
        port_def.add_group(group=port_def_group_port_spec)
        return port_def.data

    async def work(self, task_data: dict, q: asyncio.Queue) -> None:
        port_data = SensorData(sensor_id=task_data["sensorid"])
        connection_timeout = int(task_data.get("timeout", 2))
        try:
            start_time = time.time()
            connection = asyncio.open_connection(host=task_data["host"], port=int(task_data["targetport"]))
            reader, writer = await asyncio.wait_for(connection, timeout=connection_timeout)
            writer.close()
            await writer.wait_closed()
            end_time = time.time()
            response_time = (end_time - start_time) * 1000

            port_data.message = f"OK Port {task_data['targetport']} available"
            port_data.add_channel(
                name="Available",
                mode="float",
                kind="TimeResponse",
                value=float(response_time),
            )
        except (TimeoutError, asyncio.TimeoutError, ConnectionError, socket.gaierror):
            port_data.error = "Exception"
            port_data.error_code = 1
            port_data.message = "Port check failed. See log for details"
            logging.exception(
                "There has been a connection error - host:{} - port:{}".format(
                    task_data["host"], task_data["targetport"]
                )
            )

        await q.put(port_data.data)
        q.task_done()



import json
import logging
from typing import Any

import backoff
import httpx
import requests
from requests import Response
from requests.adapters import HTTPAdapter
from requests.exceptions import RetryError, ConnectionError
from urllib3.util.retry import Retry


class PRTGHTTPApi(object):
    """Represents the PRTG HTTP API.

    Attributes:
        probe_config (dict): The probe config as dict.
        server_uri (str): The URI of the PRTG Server.
        auth (frozenset): The authentication part.
        session (requests.Session): A ``requests`` session object.
        retry_strategy (urllib3.util.retry.Retry): Retry strategy for sync HTTP requests.
        adapter (requests.adapters.HTTPAdapter): The ``requests`` HTTP adapter.
        PROTOCOL_VERSION (int): Version of the communication protocol.

    """

    PROTOCOL_VERSION = 1

    def __init__(self, probe_config: dict, backoff_factor: float):
        """Constructor for the PRTGHTTPApi object

        Args:
            probe_config (dict): The probe config as dict.
            backoff_factor (float): The backoff factor for the retry strategy.

        """
        self.probe_config = probe_config
        self.server_uri = f'https://{self.probe_config["prtg_server_ip_dns"]}:{self.probe_config["prtg_server_port"]}'
        self.auth = frozenset(
            [
                ("gid", self.probe_config["probe_gid"]),
                ("key", self.probe_config["probe_access_key_hashed"]),
                ("protocol", self.PROTOCOL_VERSION),
            ]
        )
        self.session = requests.Session()
        self.retry_strategy = Retry(
            total=5,
            backoff_factor=backoff_factor,
            status_forcelist=[400, 429, 500, 502, 503, 504],
            method_whitelist=["POST"],
        )
        self.adapter = HTTPAdapter(max_retries=self.retry_strategy)
        if probe_config["disable_ssl_verification"]:
            logging.info("Disabling verification of PRTG certificate")

    def _send_api_request(self, api_url: str = None, post_data: dict = None, url_params: dict = None) -> Response:
        """Send an API request to the PRTG Core.

        Args:
            api_url (str): The API URL.
            post_data (:obj: `dict', optional): The POST data to include in the request.
            url_params (:obj: 'dict', optional): The URL parameters for the request.

        Returns:
            requests.Response: Response object for the request.

        Raises:
            RetryError: Raised when max connection retries are exceeded.
            ConnectionError: Raised when no connection is possible.

        """
        self.session.mount("https://", self.adapter)
        try:
            result = self.session.post(
                url=api_url,
                data=post_data,
                params=url_params,
                verify=not self.probe_config["disable_ssl_verification"],
            )
            if result.status_code == 401:
                logging.error("Cannot connect to the PRTG server, have you approved your probe?")
            return result
        except RetryError:
            logging.exception("Max connection retries exceeded, backing off!")
            raise
        except ConnectionError:
            logging.exception("Cannot connect to server, please check your config")
            raise

    @backoff.on_exception(
        backoff.expo,
        (httpx.TimeoutException, httpx.NetworkError, httpx.HTTPStatusError),
        max_tries=20,
        logger=logging.getLogger(),
    )
    async def _send_api_request_async(
        self, api_url: str = None, post_data: Any = None, url_params: dict = None
    ) -> Response:
        """Send an async API request to the PRTG Core.

        Args:
            api_url (str, optional): The API URL.
            post_data (:obj: `Any', optional): The POST data to include in the request.
            url_params (:obj: 'dict', optional): The URL parameters for the request.

        Returns:
            httpx._models.Response: ``httpx`` response.

        Raises:
            httpx.HTTPStatusError: Raised for status code in list.
        """
        verify_ssl = True
        if self.probe_config["disable_ssl_verification"]:
            verify_ssl = False
        async with httpx.AsyncClient(verify=verify_ssl) as client:
            request = client.build_request(method="POST", url=api_url, data=post_data, params=url_params)
            response = await client.send(request=request, timeout=10)
            if response.status_code in [400, 429, 500, 502, 503, 504]:
                raise httpx.HTTPStatusError(message="Request failed.", request=request, response=response)
            return response

    def send_announce(self, sensor_definitions: list) -> Response:
        """Send a request to the PRTG Core to announce the pyprobe.

        Args:
            sensor_definitions (:obj: `list` of :obj:`dict`): List of sensor definitions.

        Returns:
            requests.Response: The repsonse object for the announce request.

        """
        announce_url = f"{self.server_uri}/probe/announce"
        announce_params = dict(self.auth)
        announce_params["probe_base_interval"] = self.probe_config["probe_base_interval"]
        announce_params["sensors"] = json.dumps(sensor_definitions)
        announce_params["name"] = self.probe_config["probe_name"]
        logging.info(f"Sending Announce request to PRTG Core at {self.server_uri}")
        logging.debug(f"Payload for Announce request to {announce_url} is: {str(announce_params)}")
        return self._send_api_request(api_url=announce_url, post_data=announce_params)

    async def get_tasks(self) -> Response:
        """Send an async request to the PRTG Core to retrieve tasks.

        Returns:
            httpx._models.Response: ``httpx`` response.
        """
        tasks_url = f"{self.server_uri}/probe/tasks"
        logging.info(f"Sending Task request to PRTG Core at {self.server_uri}")
        logging.debug(f"Payload for Task request to {tasks_url} is: {str(dict(self.auth))}")
        return await self._send_api_request_async(api_url=tasks_url, post_data=dict(self.auth))

    async def send_data(self, sensor_response_data: list) -> Response:
        """Send an async request to the PRTG Core to submit sensor data.

        Args:
            sensor_response_data (:obj:`list` of :obj:`dict`):

        Returns:
            httpx._models.Response: ``httpx`` response.

        """
        data_url = f"{self.server_uri}/probe/data"
        logging.info(f"Sending Data Request to PRTG Core at {self.server_uri}")
        logging.debug(
            f"Payload for Data request to {data_url} with URL parameters {str(dict(self.auth))} is: "
            f"{str(sensor_response_data)}"
        )
        return await self._send_api_request_async(
            api_url=data_url,
            post_data=json.dumps(sensor_response_data),
            url_params=dict(self.auth),
        )



import logging

from logging.handlers import TimedRotatingFileHandler


def setup_logging() -> (logging.Logger, logging.StreamHandler, logging.Formatter):
    """Setup logging.

    Returns:
        tuple(logging.Logger, logging.StreamHandler, logging.Formatter): Logger, StreamHandler and Formatter.

    """
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter("%(asctime)s :: %(levelname)s :: %(message)s")

    console_handler = logging.StreamHandler()
    console_handler.setLevel(level=logging.DEBUG)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    return logger, console_handler, formatter


def setup_file_logging(
    cfg: dict, logger: logging.Logger, formatter: logging.Formatter, console_handler: logging.StreamHandler
) -> logging.Logger:
    """Set up file logging.

    Args:
        cfg (dict): ProbeConfig as dict.
        logger (logging.Logger): A valid logger.
        formatter (logging.Formatter): Formatter for logger.
        console_handler (logging.StreamHandler): StreamHandler for logger.

    Returns:
        logging.Logger: A file logger.

    Raises:
        Exception: Very broad exception.

    Todo:
        * Catch more specific exceptions.

    """
    try:
        logger.setLevel(cfg["log_level"])

        if cfg["log_file_location"]:
            filehandler = TimedRotatingFileHandler(cfg["log_file_location"], when="d", interval=1, backupCount=7)
            filehandler.setFormatter(formatter)
            logger.addHandler(filehandler)
            logger.removeHandler(console_handler)
        return logger
    # todo: catch explicit
    except Exception:
        raise



import asyncio
import logging
from typing import Any

import httpx
from prometheus_client import parser

from prtg_pyprobe.sensors.helpers import (
    SensorData,
    SensorDefinitionGroup,
    SensorDefinition,
)
from prtg_pyprobe.sensors.sensor import SensorBase


class Sensor(SensorBase):
    @property
    def name(self) -> str:
        return "Prometheus Exporter (experimental)"

    @property
    def kind(self) -> str:
        return "mpprometheusexporter"

    @property
    def definition(self) -> dict:
        prometheus_exporter_def = SensorDefinition(
            name=self.name,
            kind=self.kind,
            description="Monitors a Prometheus exporter",
            sensor_help="Monitors a Prometheus exporter and pulls select metrics.",
            tag="mpprometheusexporter",
        )
        prometheus_exporter_def_group_target = SensorDefinitionGroup(name="target_info", caption="Target Information")
        prometheus_exporter_def_group_metric = SensorDefinitionGroup(name="metric_info", caption="Metric Information")
        prometheus_exporter_def_group_target.add_field_timeout(10, 10, 900)
        prometheus_exporter_def_group_target.add_field(
            field_type="edit",
            name="port",
            caption="Exporter Port",
            required="1",
            help="Enter the port the Prometheus Exporter is listening on.",
        )
        prometheus_exporter_def_group_metric.add_field(
            field_type="edit",
            name="metrics",
            caption="Metrics",
            default="",
            required="1",
            help="Enter a list of comma separated metrics to capture.",
        )
        prometheus_exporter_def_group_metric.add_field(
            field_type="edit",
            name="metric_path",
            caption="Metric Path",
            default="/metrics",
            required="1",
            help="Enter the path to the metrics, e.g. /metrics",
        )
        prometheus_exporter_def.add_group(group=prometheus_exporter_def_group_target)
        prometheus_exporter_def.add_group(group=prometheus_exporter_def_group_metric)
        return prometheus_exporter_def.data

    async def work(self, task_data: dict, q: asyncio.Queue) -> None:
        prometheus_data = SensorData(sensor_id=task_data["sensorid"])
        try:
            uri = f"http://{task_data['host']}:{task_data['port']}{task_data['metric_path']}"

            async with httpx.AsyncClient() as client:
                request = client.build_request(
                    method="GET",
                    url=uri,
                )
                response = await client.send(request=request, timeout=int(task_data["timeout"]))

            res_txt = response.text
            metrics = parser.text_string_to_metric_families(res_txt)
            for family in metrics:
                for sample in family.samples:
                    if sample.name in task_data["metrics"]:
                        prometheus_data.add_channel(
                            name=f"{sample.name} - {sample.labels}",
                            mode="float",
                            kind="Custom",
                            customunit="",
                            value=self.cast_to_float(value=sample.value, sample_name=sample.name),
                        )
            prometheus_data.message = "OK"
        except CastException:
            prometheus_data.error = "Exception"
            prometheus_data.error_code = 1
            prometheus_data.message = "Could not cast the value returned to the correct format."
            logging.exception("Casting error.")
        except httpx.HTTPError:
            prometheus_data.error = "Exception"
            prometheus_data.error_code = 1
            prometheus_data.message = "HTTP request failed. See log for details."
            logging.exception("There was an error in httpx.")
        except KeyError:
            prometheus_data.error = "Exception"
            prometheus_data.error_code = 1
            prometheus_data.message = "Something went wrong. See logs for details."
            logging.exception("Something went wrong.")
        except ValueError:
            prometheus_data.error = "Exception"
            prometheus_data.error_code = 1
            prometheus_data.message = "Something went wrong. See logs for details."
            logging.exception("Something went wrong.")

        await q.put(prometheus_data.data)
        q.task_done()

    @staticmethod
    def cast_to_float(value: Any, sample_name: str) -> float:
        try:
            return float(value)
        except ValueError:
            raise CastException(value=value, sample_name=sample_name)


class CastException(Exception):
    def __init__(self, value: Any, sample_name: str):
        self._message = f"Could not convert {value} for sample {sample_name}"
        super().__init__(self._message)



import json
import logging
import os
import traceback
import uuid
from pathlib import Path

import regex
import yaml
from prompt_toolkit.validation import Validator, ValidationError

from prtg_pyprobe.utils.validators import (
    validation,
    validation_schemes,
    ValidationFailedError,
)


class IPDNSValidator(Validator):
    def validate(self, document):
        """Check if input is a IP address or DNS name.

        Args:
            document: ``prompt`` toolkit document.

        Raises:
            ValidationError: If the document does not match validation rule.

        """
        ip_dns_ok = regex.match(
            r"^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$|"
            r"^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)+([A-Za-z]|[A-Za-z][A-Za-z0-9\-]*[A-Za-z0-9])$",
            document.text,
        )
        if not ip_dns_ok or document.text == "":
            raise ValidationError(
                message="Please Enter a valid IP/DNS name.",
                cursor_position=len(document.text),
            )


class DirectoryWriteableValidator(Validator):
    def validate(self, document):
        """Check if input directory is writeable.

        Args:
            document: ``prompt`` toolkit document.

        Raises:
            ValidationError: If the document does not match validation rule.

        """
        if document.text:
            try:
                os.chdir(Path(document.text).parent)
                f = open(document.text, "w+")
                f.write("This is your logfile")
                f.close()
            except PermissionError:
                raise ValidationError(
                    message="Please specify a path that is writeable for the user running the probe",
                    cursor_position=len(document.text),
                )
            except IsADirectoryError:
                raise ValidationError(
                    message='Please specify a directory and filename that is writeable i.e.  "~/somefolder/probe.log"',
                    cursor_position=len(document.text),
                )
        else:
            print("Empty string specified, no logfile will be saved and logs will be printed to stdout!")


class PortRangeValidator(Validator):
    def validate(self, document):
        """Check if integer is between 1 and 65535.

        Args:
            document: ``prompt`` toolkit document.

        Raises:
            ValidationError: If the document does not match validation rule.

        """
        port_ok = True
        try:
            port_number = int(document.text)
            if not 0 < port_number < 65536:
                port_ok = False
        except ValueError:
            port_ok = False
        if not port_ok or document.text == "":
            raise ValidationError(
                message="Please enter a valid port in the range 1 - 65535.",
                cursor_position=len(document.text),
            )


class GUIDValidator(Validator):
    def validate(self, document):
        """Check if string is valid GUID.

        Args:
            document: ``prompt`` toolkit document.

        Raises:
            ValidationError: If the document does not match validation rule.

        """
        guid_ok = regex.match(
            r"([a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}){1}",
            document.text,
        )
        if not guid_ok or document.text == "":
            raise ValidationError(
                message="Probe GID has to be in GUID Format",
                cursor_position=(len(document.text)),
            )


class BaseIntervalValidator(Validator):
    def validate(self, document):
        """Check if input interval is between 60 and 300 seconds.

        Args:
            document: ``prompt`` toolkit document.

        Raises:
            ValidationError: If the document does not match validation rule,

        """
        interval_ok = True
        try:
            interval = int(document.text)
            if not 59 < interval < 301:
                interval_ok = False
        except ValueError:
            interval_ok = False
        if not interval_ok or document.text == "":
            raise ValidationError(
                message="Interval has to be between 60 and 300.",
                cursor_position=(len(document.text)),
            )


class BaseChunkSizeValidator(Validator):
    def validate(self, document):
        chunk_size_ok = True
        try:
            chunk_size = int(document.text)
            if not 5 < chunk_size < 50:
                chunk_size_ok = False
        except ValueError:
            chunk_size_ok = False
        if not chunk_size_ok or document.text == "":
            raise ValidationError(
                message="Chunk Size has to be between 5 and 50.",
                cursor_position=(len(document.text)),
            )


class NotEmptyValidator(Validator):
    def validate(self, document):
        """Check if input is empty.

        Args:
            document: ``prompt`` toolkit document.

        Raises:
            ValidationError: If the document does not match validation rule.

        """
        if document.text == "":
            raise ValidationError(message="Input cannot be empty.", cursor_position=len(document.text))


class ProbeConfig(object):
    """An object representing the probe config providing read and write functions."""

    def __init__(self, path: str):
        """Constructor of ProbeConfig.

        Args:
            path (str): The path to the config file.

        """
        self.config_path = path

    def write(self, config_data: dict):
        """Write the given config data to the path provided to the constructor.

        Args:
            config_data (dict): Config data as dict.

        """
        with open(self.config_path, "w") as f:
            yaml.dump(config_data, f)

    def read(self) -> dict:
        """Read the config from file or ENV.

        Returns:
            dict: Config as dict.

        Raises:
            ValidationFailedError: Is raised when config can not be validated.
            KeyError: Is raised when there is no config in the ENV.

        """
        try:
            with open(self.config_path, "r") as f:
                config = yaml.load(f, Loader=yaml.FullLoader)
            return config
        except FileNotFoundError:
            try:
                # todo: think about different variables instead of one which holds json
                logging.info("No config file found, attempting to read from ENV!")
                env_probe_conf = os.environ["PROBE_CONFIG"]
                try:
                    config = validation(
                        val=json.loads(env_probe_conf),
                        scheme=validation_schemes["ProbeConfig"],
                    )
                    return config
                except ValidationFailedError:
                    logging.error(traceback.format_exc())
                    logging.error("Loading of config from ENV failed. Wrong Format!")
                    raise
            except KeyError:
                logging.error("There is no config in ENV!")
                raise


configuration_questions = [
    {
        "type": "input",
        "name": "log_file_location",
        "message": "Where do you want the log file to be saved? (empty means logs will be sent to stdout)",
        "default": "/var/log/pyprobe/pyprobe.log",
        "validate": DirectoryWriteableValidator,
    },
    {
        "type": "input",
        "name": "prtg_server_ip_dns",
        "message": "IP or DNS of your PRTG Server:",
        "validate": IPDNSValidator,
    },
    {
        "type": "input",
        "name": "prtg_server_port",
        "message": "The port your PRTG server, listens for mini probe connections (only HTTPS connection allowed):",
        "default": "443",
        "validate": PortRangeValidator,
    },
    {
        "type": "input",
        "name": "probe_gid",
        "message": "Probe Gid (GUID/UUID):",
        "default": str(uuid.uuid4()).upper(),
        "validate": GUIDValidator,
    },
    {
        "type": "input",
        "name": "probe_access_key",
        "message": "The access key as defined on the PRTG Server:",
        "validate": NotEmptyValidator,
    },
    {
        "type": "input",
        "name": "probe_name",
        "message": "The name of the probe:",
        "default": "Python Mini Probe",
        "validate": NotEmptyValidator,
    },
    {
        "type": "input",
        "name": "probe_base_interval",
        "message": "The base interval for sensor in seconds (only values between 60 and 300 supported):",
        "default": "60",
        "validate": BaseIntervalValidator,
    },
    {
        "type": "input",
        "name": "probe_task_chunk_size",
        "message": "The task chunk size, default 20, powerful machines may be able to handle more "
        "(only values between 5 and 50 supported):",
        "default": "20",
        "validate": BaseChunkSizeValidator,
    },
    {
        "type": "list",
        "name": "log_level",
        "message": "What level of logging do you want?",
        "choices": ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"],
    },
    {
        "type": "confirm",
        "message": "Do you want to turn off Certificate Validation for connections to the PRTG Core? "
        "(recommended when using a self signed certificate)",
        "name": "disable_ssl_verification",
        "default": True,
    },
]



import asyncio
import logging

import psutil

from prtg_pyprobe.sensors.helpers import (
    SensorProbeData,
    SensorDefinition,
)
from prtg_pyprobe.sensors.sensor import SensorPSUtilBase


class Sensor(SensorPSUtilBase):
    @property
    def name(self) -> str:
        return "CPU Load (Probe)"

    @property
    def kind(self) -> str:
        return "mpcpuload"

    @property
    def definition(self) -> dict:
        cpu_load_def = SensorDefinition(
            name=self.name,
            kind=self.kind,
            description="Monitors CPU load avg on the system the pyprobe is running on.",
            sensor_help="Monitors CPU load avg on the system the pyprobe is running on."
            "This is always the case even if added to a different device than the Probe Device.",
            tag="mpprobehealthsensor",
        )
        return cpu_load_def.data

    async def work(self, task_data: dict, q: asyncio.Queue) -> None:
        cpu_load_data = SensorProbeData(sensor_id=task_data["sensorid"])
        try:
            cpu_load = self.get_cpu_load()
            cpu_load_data.add_load_avg_channel(cpu_load[0], 1)
            cpu_load_data.add_load_avg_channel(cpu_load[1], 5)
            cpu_load_data.add_load_avg_channel(cpu_load[2], 10)
            cpu_load_data.message = "OK"
        except psutil.Error:
            cpu_load_data.error = "Exception"
            cpu_load_data.error_code = 1
            cpu_load_data.message = "CPU Load check failed. See log for details"
            logging.exception("CPU Load avg could not be determined")
        await q.put(cpu_load_data.data)
        q.task_done()



import asyncio
import logging
import platform
import time

import psutil

from prtg_pyprobe.sensors.helpers import (
    SensorProbeData,
    SensorDefinition,
)
from prtg_pyprobe.sensors.sensor import SensorPSUtilBase


class Sensor(SensorPSUtilBase):
    @property
    def name(self) -> str:
        return "Probe Health"

    @property
    def kind(self) -> str:
        return "mpprobehealth"

    @property
    def definition(self) -> dict:
        probe_health_def = SensorDefinition(
            name=self.name,
            kind=self.kind,
            description="Sensor used to monitor the health of the system on which the python probe is runnning",
            default="yes",
            sensor_help="Sensor used to monitor the health of the system on which the python probe is runnning",
            tag="mpprobehealthsensor",
        )
        probe_health_def.add_group(self.temperature_settings_group)
        probe_health_def.add_group(self.diskspace_settings_group)
        return probe_health_def.data

    async def work(self, task_data: dict, q: asyncio.Queue) -> None:
        probe_health_data = SensorProbeData(sensor_id=task_data["sensorid"])
        try:
            uname = platform.uname()
            probe_health_data.message = f"Running on {uname.system} in version {uname.version}! OK"
            cpu_load = psutil.getloadavg()
            vmemory, swapmemory = self.get_memory_usage()
            disk_partitions = psutil.disk_partitions()
            sys_uptime = psutil.boot_time()

            temperatures = self.get_system_temperatures(fahrenheit=True if task_data["celfar"] == 2 else False)
            if temperatures:
                for name, entries in temperatures.items():
                    for entry in entries:
                        probe_health_data.add_temperature_channel(entry.label or name, entry.current)

            for partition in disk_partitions:
                part_disk_usage = self.get_partition_usage(partition)
                probe_health_data.add_disk_space_percentage_use(partition, part_disk_usage)

                if task_data["diskfull"] == "2":
                    probe_health_data.add_disk_space_details_channel("Total", partition, part_disk_usage[0])
                    probe_health_data.add_disk_space_details_channel("Used", partition, part_disk_usage[1])
                    probe_health_data.add_disk_space_details_channel("Free", partition, part_disk_usage[2])

            probe_health_data.add_channel(name="Memory Total", mode="integer", kind="BytesMemory", value=vmemory.total)
            probe_health_data.add_channel(
                name="Memory Available",
                mode="integer",
                kind="BytesMemory",
                value=vmemory.available,
            )
            probe_health_data.add_channel(
                name="Swap Total",
                mode="integer",
                kind="BytesMemory",
                value=swapmemory.total,
            )
            probe_health_data.add_channel(name="Swap Free", mode="integer", kind="BytesMemory", value=swapmemory.free)

            probe_health_data.add_load_avg_channel(cpu_load[0], 1)
            probe_health_data.add_load_avg_channel(cpu_load[1], 5)
            probe_health_data.add_load_avg_channel(cpu_load[2], 10)

            probe_health_data.add_channel(
                name="System Uptime", mode="float", kind="TimeSeconds", value=float(time.time() - sys_uptime)
            )
        except psutil.Error:
            probe_health_data.error = "Exception"
            probe_health_data.error_code = 1
            probe_health_data.message = "Probehealth check failed. See log for details"
            logging.exception("Probe health information could not be determined")
        await q.put(probe_health_data.data)
        q.task_done()



import asyncio
import logging

import psutil

from prtg_pyprobe.sensors.helpers import (
    SensorData,
    SensorDefinitionGroup,
    SensorDefinition,
)
from prtg_pyprobe.sensors.sensor import SensorBase


class Sensor(SensorBase):
    @property
    def name(self) -> str:
        return "Traffic (Probe)"

    @property
    def kind(self) -> str:
        return "mptraffic"

    @property
    def definition(self) -> dict:
        traffic_def = SensorDefinition(
            name=self.name,
            kind=self.kind,
            description="Monitors traffic on the host the probe is running on.",
            sensor_help="Monitors traffic on the host the probe is running on. "
            "This is always the case even if added to a different device than the Probe Device.",
            tag="mptrafficsensor",
        )
        traffic_def_group_sensor_specific = SensorDefinitionGroup(name="sensor_specific", caption="Sensor Specific")
        traffic_def_group_sensor_specific.add_field(
            field_type="edit",
            name="iface_name",
            caption="Interface Name",
            required="1",
            help="Provide the interface name to monitor.",
        )
        traffic_def.add_group(traffic_def_group_sensor_specific)
        return traffic_def.data

    async def work(self, task_data: dict, q: asyncio.Queue) -> None:
        traffic_data = SensorData(sensor_id=task_data["sensorid"])
        try:
            traffic_data.message = f"Interface {task_data['iface_name']}"
            traffic_total = psutil.net_io_counters(pernic=True)
            iface_traffic = traffic_total[task_data["iface_name"]]
            traffic_data.add_channel(
                name="Traffic out",
                mode="counter",
                kind="BytesBandwidth",
                value=iface_traffic.bytes_sent,
            )
            traffic_data.add_channel(
                name="Traffic in",
                mode="counter",
                kind="BytesBandwidth",
                value=iface_traffic.bytes_recv,
            )
            traffic_data.add_channel(
                name="Errors in",
                mode="integer",
                kind="Custom",
                customunit="",
                value=iface_traffic.errin,
            )
            traffic_data.add_channel(
                name="Errors out",
                mode="integer",
                kind="Custom",
                customunit="",
                value=iface_traffic.errout,
            )
            traffic_data.add_channel(
                name="Dropped in",
                mode="integer",
                kind="Custom",
                customunit="",
                value=iface_traffic.dropin,
            )
            traffic_data.add_channel(
                name="Dropped out",
                mode="integer",
                kind="Custom",
                customunit="",
                value=iface_traffic.dropout,
            )
        except KeyError:
            traffic_data.error_code = 1
            traffic_data.error = "Exception"
            traffic_data.message = f"Interface {task_data['iface_name']} not found"
            logging.exception("Interface not found.")
        await q.put(traffic_data.data)
        q.task_done()



import asyncio
import json
import time

import pytest
from pysnmp.hlapi.asyncio import SnmpEngine
from requests import Response
from testfixtures import LogCapture

import prtg_pyprobe.monitoring
import prtg_pyprobe.run
from prtg_pyprobe.communication.communication_http_api import PRTGHTTPApi
from prtg_pyprobe.monitoring import monitoring
from prtg_pyprobe.run import main, monitoring_loop


def return_response_200(*args, **kwargs):
    resp = Response()
    resp.status_code = 200
    resp._content = (
        b'{"sensorid": "1234", "timeout": 5, "target": "https://example.com","request_type": "GET",'
        b'"acceptable_status_codes": "","headers": "","post_data": "",}'
    )
    return resp


async def return_async_response_200(*args, **kwargs):
    resp = Response()
    resp.status_code = 200
    resp._content = (
        '{"sensorid": "1234", "timeout": 5, "target": "https://example.com", "request_type": "GET", '
        '"acceptable_status_codes": "", "headers": "", "post_data": ""}'.encode("utf-8")
    )
    return resp


def return_dict_json(*args, **kwargs):
    return [
        {
            "sensorid": "1234",
            "timeout": 5,
            "target": "https://example.com",
            "request_type": "GET",
            "acceptable_status_codes": "",
            "headers": "",
            "post_data": "",
        }
    ]


class MockEventLoop(object):
    @staticmethod
    def run_until_complete(self):
        return "work"

    @staticmethod
    def is_running():
        return False


def mock_event_loop():
    return MockEventLoop()


async def return_true_async(*args, **kwargs):
    return True


class TestRun:
    @pytest.mark.asyncio
    async def test_monitoring_loop(self, monkeypatch, prtg_api, probe_config_dict):
        monkeypatch.setattr(PRTGHTTPApi, "get_tasks", return_async_response_200)
        monkeypatch.setattr(PRTGHTTPApi, "send_data", return_async_response_200)
        monkeypatch.setattr(asyncio, "sleep", return_true_async)
        monkeypatch.setattr(monitoring, "monitoring", return_true_async)
        with LogCapture() as log:
            await monitoring_loop(
                prtg_api=prtg_api, sensors=[], run_monitoring=True, snmp_eng=SnmpEngine(), config=probe_config_dict
            )
        log.check_present(
            (
                "root",
                "DEBUG",
                "Tasks received from PRTG {'sensorid': '1234', 'timeout': 5, 'target': 'https://example.com', "
                "'request_type': 'GET', 'acceptable_status_codes': '', 'headers': '', 'post_data': ''}",
            )
        )

    def test_run(self, monkeypatch, mocker):
        monkeypatch.setenv(
            name="PROBE_CONFIG",
            value='{"disable_ssl_verification": false, "log_file_location": "", "log_level": "DEBUG",'
            ' "probe_access_key": '
            '"miniprobe", "probe_access_key_hashed": "cd7b773e2ce4205e9f5907b157f3d26495c5b373", '
            '"probe_base_interval": "60", "probe_task_chunk_size": "20", '
            '"probe_gid": "72C461B5-F768-470B-A1A8-2D5F5DEDDF8F", '
            '"probe_name": "Python Mini Probe UT", "probe_protocol_version": "1", '
            '"prtg_server_ip_dns": "ut.prtg,com", "prtg_server_port": "443"}',
        )
        config = {
            "disable_ssl_verification": False,
            "log_file_location": "",
            "log_level": "DEBUG",
            "probe_access_key": "miniprobe",
            "probe_access_key_hashed": "cd7b773e2ce4205e9f5907b157f3d26495c5b373",
            "probe_base_interval": "60",
            "probe_task_chunk_size": "20",
            "probe_gid": "72C461B5-F768-470B-A1A8-2D5F5DEDDF8F",
            "probe_name": "Python Mini Probe UT",
            "probe_protocol_version": "1",
            "prtg_server_ip_dns": "ut.prtg,com",
            "prtg_server_port": "443",
        }
        monkeypatch.setattr(PRTGHTTPApi, "send_announce", return_response_200)
        monkeypatch.setattr(Response, "json", return_dict_json)
        monkeypatch.setattr(Response, "json", return_dict_json)
        monkeypatch.setattr(asyncio, "get_event_loop", mock_event_loop)
        monkeypatch.setattr(prtg_pyprobe.monitoring, "monitoring", True)
        monkeypatch.setattr(time, "sleep", True)
        with LogCapture() as log:
            main()
        log.check_present(
            ("root", "INFO", "Starting pyprobe!"),
            ("root", "INFO", "Loading configuration"),
            ("root", "INFO", "Config loaded and probe announcing to PRTG"),
            ("root", "DEBUG", f"Your current config: \n {json.dumps(config, indent=2)}"),
            ("root", "INFO", "Initializing SNMP Engine"),
            ("root", "INFO", "Initializing PRTG API"),
            ("root", "INFO", "Initializing Sensors"),
            ("root", "INFO", "Init done!"),
        )



import asyncio
import logging
import time

from pysnmp.error import PySnmpError
from pysnmp.hlapi.asyncio import SnmpEngine, UsmUserData, CommunityData
from pysnmp.smi.error import NoSuchObjectError

from prtg_pyprobe.sensors.helpers import SensorData, SensorDefinition
from prtg_pyprobe.sensors.sensor import SensorSNMPBase


class Sensor(SensorSNMPBase):
    @property
    def name(self) -> str:
        return "SNMP CPU Load Avg"

    @property
    def kind(self) -> str:
        return "mpsnmpcpuloadavg"

    @property
    def definition(self) -> dict:
        snmp_cpu_load_def = SensorDefinition(
            name=self.name,
            kind=self.kind,
            description="Monitors Load Avg via SNMP.",
            sensor_help="This sensor monitors the load avg of a given target via SNMP.",
            tag="mpsnmpcpuloadavg",
        )
        snmp_cpu_load_def.add_group(self.snmp_specific_settings_group)
        return snmp_cpu_load_def.data

    async def work(self, task_data: dict, q: asyncio.Queue, snmp_engine: SnmpEngine):
        snmp_cpu_load_data = SensorData(sensor_id=task_data["sensorid"])
        creds = UsmUserData(userName=task_data["community"])
        if task_data["snmp_version"] == "0" or task_data["snmp_version"] == "1":
            creds = CommunityData(task_data["community"], mpModel=int(task_data["snmp_version"]))
        try:
            start_time = time.time()
            results = await self.get(
                target=task_data["host"],
                port=int(task_data["port"]),
                credentials=creds,
                oids=[
                    ".1.3.6.1.4.1.2021.10.1.3.1",
                    ".1.3.6.1.4.1.2021.10.1.3.2",
                    ".1.3.6.1.4.1.2021.10.1.3.3",
                ],
                engine=snmp_engine,
            )
            end_time = time.time()
            response_time = (end_time - start_time) * 1000

            for result in results:
                k, v = str(result[0]), self.cast(result[1])
                if v == "":
                    raise PySnmpError("Returned value empty. Probably no such value/object")
                if "1.3.1" in k:
                    k = "Load Avg 1 min"
                elif "1.3.2" in k:
                    k = "Load Avg 5min"
                elif "1.3.3" in k:
                    k = "Load Avg 10 min"
                snmp_cpu_load_data.add_channel(
                    name=k,
                    mode="float",
                    unit="Custom",
                    customunit="",
                    value=v,
                )
            snmp_cpu_load_data.add_channel(
                name="Response Time",
                mode="float",
                kind="TimeResponse",
                value=float(response_time),
            )
            snmp_cpu_load_data.message = "OK"
        except NoSuchObjectError:
            snmp_cpu_load_data.error = "Exception"
            snmp_cpu_load_data.error_code = 1
            snmp_cpu_load_data.message = "The OID is wrong."
            logging.exception("No such object. Probably the given OID is wrong.")
        except PySnmpError:
            snmp_cpu_load_data.error = "Exception"
            snmp_cpu_load_data.error_code = 1
            snmp_cpu_load_data.message = "SNMP Query failed. See logs for details"
            logging.exception("There has been an error in PySNMP.")
        except RuntimeError:
            snmp_cpu_load_data.error = "Exception"
            snmp_cpu_load_data.error_code = 1
            snmp_cpu_load_data.message = "SNMP Query failed. See logs for details"
            logging.exception("There has been an error.")
        except ValueError:
            snmp_cpu_load_data.error = "Exception"
            snmp_cpu_load_data.error_code = 1
            snmp_cpu_load_data.message = "The SNMP Query did return a not supported value type."
            logging.exception("The SNMP query probably returned an unsupported value.")
        await q.put(snmp_cpu_load_data.data)
        q.task_done()



from setuptools import setup

setup(
    name="prtg_pyprobe",
    description="Python Probe for PRTG Network Monitor",
    package_data={"": ["*.service"]},
    install_requires=[
        "click",
        "requests",
        "PyInquirer",
        "pythonping",
        "pyyaml",
        "certifi",
        "cerberus",
        "psutil",
        "aiodns",
        "cchardet",
        "dnspython",
        "pysnmp",
        "aioping",
        "httpx",
        "python-daemon",
        "backoff",
        "aioboto3",
        "dpath",
        "prometheus_client",
    ],
    tests_require=[
        "pytest",
        "pytest-cov",
        "pytest-asyncio",
        "pytest-mock",
        "pytest-subprocess",
        "black",
        "pyfakefs",
        "testfixtures",
        "flake8",
        "asynctest",
    ],
)



import asyncio
import json
import logging
import sys

from pysnmp.hlapi.asyncio import SnmpEngine

from prtg_pyprobe.communication.communication_http_api import PRTGHTTPApi
from prtg_pyprobe.monitoring import monitoring
from prtg_pyprobe.utils.config import ProbeConfig
from prtg_pyprobe.utils.defaults import CONFIG_PATH
from prtg_pyprobe.utils.logging import setup_logging, setup_file_logging
from prtg_pyprobe.utils.sensor_loader import create_sensor_objects


async def monitoring_loop(
    prtg_api: PRTGHTTPApi, sensors: list, snmp_eng: SnmpEngine, run_monitoring: bool, config: dict
):
    """The monitoring functionality of the pyprobe.

    Args:
        prtg_api (PRTGHTTPApi): PRTGHTTPApi object.
        sensors (list): List of sensor objects.
        snmp_eng (SnmpEngine): SNMP Engine object.
        run_monitoring (bool): Bool which determines whether to run monitoring or not.
        config (dict): Probe configuration as dict.

    """
    while run_monitoring:
        tasks_api_response = await prtg_api.get_tasks()
        if tasks_api_response.status_code == 200:
            tasks = tasks_api_response.json()
            logging.debug(f"Tasks received from PRTG {tasks}")
            if len(tasks) > 0:
                await monitoring.monitoring(
                    tasks_list=tasks,
                    sensor_objects=sensors,
                    snmp_engine=snmp_eng,
                    prtg_api=prtg_api,
                    config=config,
                )
        if tasks_api_response.status_code == 401:
            logging.warning("Please check that your probe has been approved in PRTG!")
        if await asyncio.sleep(10):
            break


def main():
    """The pyprobe main function."""
    logger, console_handler, formatter = setup_logging()
    logger.info("Starting pyprobe!")
    try:
        logger.info("Loading configuration")
        cfg = ProbeConfig(path=f"{CONFIG_PATH}config.yml").read()
    except FileNotFoundError:
        logger.exception("No configuration file found, ending.")
        sys.exit(1)

    # noinspection PyBroadException
    try:
        if not cfg["log_file_location"] == "":
            setup_file_logging(cfg, logger, formatter, console_handler)
            logger.info("Now logging to your logfile")
        logger.setLevel(cfg["log_level"])
    except Exception:
        logger.exception("Something bad happened!")
        sys.exit(1)

    logger.info("Config loaded and probe announcing to PRTG")
    logger.debug(f"Your current config: \n {json.dumps(cfg, indent=2)}")

    logger.info("Initializing SNMP Engine")
    snmp_eng = SnmpEngine()

    logger.info("Initializing PRTG API")
    prtg_api = PRTGHTTPApi(probe_config=cfg, backoff_factor=10)
    logger.info("Initializing Sensors")
    sensors = create_sensor_objects()
    sensor_defs = [sensor.definition for sensor in sensors]
    logger.info("Init done!")
    send_announce = True
    run_monitoring = False
    while send_announce:
        announce_api_response = prtg_api.send_announce(sensor_definitions=sensor_defs)
        if announce_api_response.status_code == 200:
            send_announce = False
            run_monitoring = True
    monitoring_event_loop = asyncio.get_event_loop()
    monitoring_event_loop.run_until_complete(
        monitoring_loop(
            prtg_api=prtg_api,
            sensors=sensors,
            snmp_eng=snmp_eng,
            run_monitoring=run_monitoring,
            config=cfg,
        )
    )


if __name__ == "__main__":
    main()



import asyncio
import logging
import time

import dns.asyncresolver
from dns.exception import DNSException

from prtg_pyprobe.sensors.helpers import (
    SensorData,
    SensorDefinitionGroup,
    SensorDefinition,
)
from prtg_pyprobe.sensors.sensor import SensorBase


class NoMatchException(Exception):
    def __init__(self, string_filter, sensor_last_message):
        self.string_filter = string_filter
        self.last_message = sensor_last_message


class Sensor(SensorBase):
    @property
    def name(self) -> str:
        return "DNS"

    @property
    def kind(self) -> str:
        return "mpdns"

    @property
    def definition(self) -> dict:
        dns_def = SensorDefinition(
            name=self.name,
            kind=self.kind,
            description="Monitors a DNS server (Domain Name Service), resolves a domain name, "
            "and compares it to an IP address",
            sensor_help="The DNS sensor monitors a Domain Name Service (DNS) server. "
            "It resolves a domain name and compares it to a given IP address.",
            tag="mpdnssensor",
        )
        dns_def_group_dns_spec = SensorDefinitionGroup(name="dns_specific", caption="DNS Specific")
        dns_def_group_dns_spec.add_field_timeout(5, 1, 900)
        dns_def_group_dns_spec.add_field(
            field_type="integer",
            name="port",
            caption="Port",
            required="1",
            default=53,
            minimum=1,
            maximum=65535,
            help="Enter the port on which the DNS service of the parent device is running.",
        )
        dns_def_group_dns_spec.add_field(
            field_type="edit",
            name="domain",
            caption="Domain",
            required="1",
            help="Enter a DNS name to resolve.",
        )
        dns_def_group_dns_spec.add_field(
            field_type="edit",
            name="filter",
            caption="Filter",
            help="Enter comma separated strings to look for in the result. i.e. 192.168.1.0,192.168.2.0 or "
            "MX=10 mx.example.com,MX=20 mx2.example.com.  "
            "These values will be compared to the last message shown in the sensor."
            " If it doesn't match the resolved record, the sensor will "
            "go into an error state, if left blank the dns record will not be checked",
        )
        dns_def_group_dns_spec.add_field(
            field_type="radio",
            name="type",
            caption="Query Type",
            required="1",
            help="Specify the type of query that the sensor will send to the DNS server.",
            options={
                "A": "Host address IPv4 (A)",
                "AAAA": "Host address IPv6 (AAAA)",
                "CNAME": "Canonical name for an alias (CNAME)",
                "MX": "Mail exchange (MX)",
                "NS": "Authoritative name server (NS)",
                "SOA": "Start of a zone of authority marker (SOA)",
                "SRV": "Service Record",
                "TXT": "Text Record",
            },
            default="A",
        )
        dns_def.add_group(dns_def_group_dns_spec)
        return dns_def.data

    async def work(self, task_data: dict, q: asyncio.Queue) -> None:
        dns_data = SensorData(sensor_id=task_data["sensorid"])
        try:
            start = time.time()
            resolver = dns.asyncresolver.Resolver()
            resolver.nameservers = [f"{task_data['host']}"]
            resolver.port = int(task_data["port"])
            res = await resolver.resolve(task_data["domain"], rdtype=task_data["type"])

            end = (time.time() - start) * 1000
            msg = f"Query Type: {task_data['type']} for {task_data['domain']}. Result: "
            if task_data["type"] == "A" or task_data["type"] == "AAAA":
                for rdata in res:
                    msg += f"{str(rdata.address)}, "
            elif task_data["type"] == "MX":
                for rdata in res:
                    msg += f"{rdata.preference}: {rdata.exchange}"
            elif task_data["type"] == "SOA":
                for rdata in res:
                    msg += (
                        f"NS: {rdata.mname}, TECH: {rdata.rname}, S/N: {rdata.serial}, "
                        f"Refresh: {rdata.refresh / 60} min, Expire: {rdata.expire / 60} min"
                    )
            elif task_data["type"] == "TXT":
                for rdata in res:
                    for string in rdata.strings:
                        msg += f"{string.decode('utf-8') }"
            else:
                for rdata in res:
                    msg += f"{rdata.target}, "
            dns_data.message = msg
            split_strings = task_data["filter"].split(",")
            if task_data["filter"]:
                for string in split_strings:
                    if string not in msg:
                        logging.debug(f'{msg} did not contain filtered string "{string}"')
                        raise NoMatchException(string_filter=string, sensor_last_message=msg)
            dns_data.add_channel(
                name="Repsonse Time",
                mode="float",
                kind="TimeResponse",
                value=float(end),
            )
        except DNSException:
            dns_data.error = "Exception"
            dns_data.error_code = 1
            dns_data.message = "DNS check failed. See log for details"
            logging.exception("Error in dnspython.")
        except NoMatchException as e:
            dns_data.error = "Exception"
            dns_data.error_code = 1
            dns_data.message = (
                f"Filtered value '{e.string_filter}' not found in last message '{e.last_message}'. See log for details"
            )
            logging.exception("Filter did not match!")
        await q.put(dns_data.data)
        q.task_done()



import asyncio
import logging

import httpx

from prtg_pyprobe.sensors.helpers import (
    SensorData,
    SensorDefinitionGroup,
    SensorDefinition,
)
from prtg_pyprobe.sensors.sensor import SensorBase


class Sensor(SensorBase):
    @property
    def name(self) -> str:
        return "External IP (Probe)"

    @property
    def kind(self) -> str:
        return "mpexternalip"

    @property
    def definition(self) -> dict:
        ext_ip_def = SensorDefinition(
            name=self.name,
            kind=self.kind,
            description="Returns the external ip address of the probe.",
            sensor_help="Returns the external ip address of the probe using the website checkip.amazonaws.com."
            "This is always the case even if added to a different device than the Probe Device.",
            tag="mpexternalipsensor",
        )
        ext_ip_def_settings = SensorDefinitionGroup(name="extipsettings", caption="External IP Settings")
        ext_ip_def_settings.add_field_timeout(30, 10, 300)
        ext_ip_def.add_group(group=ext_ip_def_settings)

        return ext_ip_def.data

    async def work(self, task_data: dict, q: asyncio.Queue) -> None:
        ext_ip_data = SensorData(sensor_id=task_data["sensorid"])
        try:
            request_type = "GET"
            headers = None
            post_data = None
            url = "https://checkip.amazonaws.com/"

            async with httpx.AsyncClient() as client:
                request = client.build_request(
                    method=request_type,
                    url=url,
                    headers=headers,
                    data=post_data,
                )
                response = await client.send(request=request, timeout=int(task_data["timeout"]))

            ext_ip_data.add_channel(
                name="Response Time",
                mode="float",
                kind="TimeResponse",
                value=float(response.elapsed.microseconds / 1000),
            )
            response_text = response.text.replace("\n", "")
            ext_ip_data.message = f"Returned external IP: {response_text}"

        except httpx.HTTPError:
            ext_ip_data.error = "Exception"
            ext_ip_data.error_code = 1
            ext_ip_data.message = "External IP sensor failed. See log for details"
            logging.exception("There was an error in httpx.")

        await q.put(ext_ip_data.data)
        q.task_done()



import asyncio
import logging
import time

from pysnmp.error import PySnmpError
from pysnmp.hlapi.asyncio import SnmpEngine, UsmUserData, CommunityData

from prtg_pyprobe.sensors.helpers import (
    SensorData,
    SensorDefinitionGroup,
    SensorDefinition,
)
from prtg_pyprobe.sensors.sensor import SensorSNMPBase


class Sensor(SensorSNMPBase):
    @property
    def name(self) -> str:
        return "SNMP Traffic"

    @property
    def kind(self) -> str:
        return "mpsnmptraffic"

    @property
    def definition(self) -> dict:
        snmp_traffic_def = SensorDefinition(
            name=self.name,
            kind=self.kind,
            description="Monitors Traffic on provided interface using SNMP",
            sensor_help="Monitors Traffic on provided interface using SNMP",
            tag="mpsnmptrafficsensor",
        )
        snmp_traffic_def_group_sensor_specific = SensorDefinitionGroup(
            name="sensor_specific", caption="Sensor Specific"
        )
        snmp_traffic_def_group_sensor_specific.add_field(
            field_type="edit",
            name="ifindex",
            caption="Interface Index (ifindex)",
            required="1",
            help="Please enter the ifIndex of the interface to be monitored.",
        )
        snmp_traffic_def.add_group(snmp_traffic_def_group_sensor_specific)
        snmp_traffic_def.add_group(self.snmp_specific_settings_group)
        return snmp_traffic_def.data

    async def work(self, task_data: dict, q: asyncio.Queue, snmp_engine: SnmpEngine):
        snmp_traffic_data = SensorData(sensor_id=task_data["sensorid"])
        creds = UsmUserData(userName=task_data["community"])
        if task_data["snmp_version"] == "0" or task_data["snmp_version"] == "1":
            creds = CommunityData(task_data["community"], mpModel=int(task_data["snmp_version"]))
        try:
            oids = [f"1.3.6.1.2.1.31.1.1.1.6.{task_data['ifindex']}", f"1.3.6.1.2.1.31.1.1.1.10.{task_data['ifindex']}"]
            if task_data["snmp_version"] == "0":
                oids = [f"1.3.6.1.2.1.2.2.1.10.{task_data['ifindex']}", f"1.3.6.1.2.1.2.2.1.16.{task_data['ifindex']}"]
            start_time = time.time()
            results = await self.get(
                target=task_data["host"],
                port=int(task_data["port"]),
                credentials=creds,
                oids=oids,
                engine=snmp_engine,
            )
            end_time = time.time()
            response_time = (end_time - start_time) * 1000
            for result in results:
                if self.cast(result[1]) == "":
                    raise PySnmpError("Returned value empty. Probably no such value/object")
            snmp_traffic_data.message = f"Interface with ifindex {task_data['ifindex']}. OK!"
            snmp_traffic_data.add_channel(
                name="Traffic In", mode="counter", unit="BytesBandwidth", value=self.cast(results[0][1])
            )
            snmp_traffic_data.add_channel(
                name="Traffic Out", mode="counter", unit="BytesBandwidth", value=self.cast(results[1][1])
            )
            snmp_traffic_data.add_channel(
                name="Traffic Total",
                mode="counter",
                unit="BytesBandwidth",
                value=self.cast(int(results[1][1]) + int(results[0][1])),
            )
            snmp_traffic_data.add_channel(
                name="Response Time",
                mode="float",
                kind="TimeResponse",
                value=float(response_time),
            )
        except PySnmpError:
            snmp_traffic_data.error = "Exception"
            snmp_traffic_data.error_code = 1
            snmp_traffic_data.message = "SNMP Query failed. See logs for details"
            logging.exception("There has been an error in PySNMP.")
        await q.put(snmp_traffic_data.data)
        q.task_done()



import logging
from typing import Any

from prtg_pyprobe.utils.validators import validate


class SensorDefinitionGroup(object):
    """Represents a group within the sensor definition"""

    class SensorDefinitionGroupField(object):
        """Represents a field within a sensor definition group"""

        def __init__(self, field_type: str, name: str, caption: str, **kwargs):
            """
            Constructor of SensorDefinitionGroupField
            Args:
                field_type (str): The field type.
                name (str): The field name.
                caption (str): The field caption.
                **kwargs: Arbitrary keyword arguments. See utils.validators for details.

            """
            self._field_type = field_type
            self._name = name
            self._caption = caption
            self._kwargs = kwargs

        @property
        @validate
        def data(self) -> dict:
            """Validated data property of SensorDefinitionGroupField
            Returns:
                dict: A validated dictionary.

            """
            field = {
                "type": self._field_type,
                "name": self._name,
                "caption": self._caption,
            }
            for k, v in self._kwargs.items():
                field[k] = v
            return field

    def __init__(self, name: str, caption: str):
        """Constructor of SensorDefinitionGroup

        Args:
            name (str): The name of the group.
            caption (str):  The caption of the group.
        """
        self._name = name
        self._caption = caption
        self._fields = []

    @property
    @validate
    def data(self) -> dict:
        """Validated data property of SensorDefinitionGroup

        Returns:
            dict: A validated dictionary.

        """
        group = {"name": self._name, "caption": self._caption, "fields": self._fields}
        return group

    def add_field_timeout(self, default: int, minimum: int, maximum: int):
        """Add a timeout field to a group.

        Args:
            default (int): The default timeout.
            minimum (int): The maximum value.
            maximum (int): The minimum value.

        """
        timeout = self.SensorDefinitionGroupField(
            field_type="integer",
            name="timeout",
            caption="Timeout (in s)",
            required="1",
            default=default,
            minimum=minimum,
            maximum=maximum,
            help=f"If the reply takes longer than this value the request is aborted "
            f"and an error message is triggered. Maximum value is {maximum} sec. (= {maximum / 60} minutes)",
        )
        self._fields.append(timeout.data)

    def add_field(self, field_type: str, name: str, caption: str, **kwargs):
        """Add a field to a group.

        Args:
            field_type (str): The field type.
                The field type should be of these types: "edit", "integer", "password", "radio"
            name (str): The field name.
            caption (str): The field caption.
            **kwargs: Arbitrary keyword arguments. See utils.validators for details.

        Examples:
                >>> ping_def_group_settings.add_field(
                        field_type="integer",
                        name="ping_count",
                        caption="Ping Count",
                        required="1",
                        default=4,
                        minimum=1,
                        maximum=20,
                        help="Enter the count of Ping requests PRTG will send to the device during an interval",
                    )

        """
        field = self.SensorDefinitionGroupField(field_type=field_type, name=name, caption=caption, **kwargs)
        self._fields.append(field.data)


class SensorDefinition(object):
    """Represents the Sensor Definition."""

    class NotASensorDefinitionGroupError(Exception):
        """Exception when an added group is not of type SensorDefinitionGroup."""

        def __init__(self):
            super().__init__("Group has to be of type SensorDefinitionGroup.")

    def __init__(
        self,
        kind: str,
        name: str,
        description: str,
        sensor_help: str,
        tag: str,
        default: str = None,
    ):
        """Constructor for `SensorDefinition`.

        Args:
            kind (str): The sensor kind.
            name (str):  The sensor name.
            description (str): The sensor description.
            sensor_help (str): The sensor help. Shows when hovering over the question mark in PRTG interface.
            tag (str): The tag for the sensor.
            default (str, optional): Determines if the sensor is created when the probe is connected.
                Allowed values "yes", "1", "true".
        """
        self._kind = kind
        self._name = name
        self._description = description
        self._help = sensor_help
        self._tag = tag
        self._default = default
        self._groups = []

    @property
    @validate
    def data(self) -> dict:
        """Validated data property of `SensorDefinition`.

        Returns:
            dict: A validated dict.
        """
        definition = {
            "name": self._name,
            "kind": self._kind,
            "description": self._description,
            "help": self._help,
            "tag": self._tag,
            "groups": self._groups,
        }
        if self._default:
            definition["default"] = self._default
        logging.debug(f"Sensor Definition for sensor {self._name}: {str(definition)}")
        return definition

    def add_group(self, group: SensorDefinitionGroup):
        """Add a group to the `SensorDefinition`.

        Args:
            group (SensorDefinitionGroup):

        Raises:
            NotASensorDefinitionGroupError: Raised when the object passed is not of type `SensorDefinitionGroup`.

        """
        if isinstance(group, SensorDefinitionGroup):
            self._groups.append(group.data)
        else:
            raise self.NotASensorDefinitionGroupError


class SensorData(object):
    """Represents the SensorData."""

    class SensorDataChannel(object):
        """Represents a SensorDataChannel"""

        def __init__(self, name: str, mode: str, value: Any, **kwargs):
            """Constructor of SensorDataChannel

            Args:
                name (str): The channel name.
                mode (str): The channel mode.
                    Allowed values are "integer", "float", "counter".
                value (Any): The channel value.
                    Allowed values "integer", "float"
                **kwargs: Arbitrary keyword arguments. See utils.validators for details.

            """
            self._name = name
            self._mode = mode
            self._value = value
            self._kwargs = kwargs

        @property
        @validate
        def data(self) -> dict:
            """Validated data property of `SensorDataChannel`.

            Returns:
                dict: A validated dict.

            """
            channel = {"name": self._name, "mode": self._mode, "value": self._value}
            for k, v in self._kwargs.items():
                channel[k] = v
            return channel

    def __init__(self, sensor_id: str):
        """Constructor of SensorData

        Args:
            sensor_id (str): The sensor id. Contained in the taskdata from the PRTG Core.
        """
        self._sensor_id = sensor_id
        self._message = None
        self._error = None
        self._error_code = None
        self._channel_list = []

    @property
    def message(self) -> str:
        """str: Property message of SensorData"""
        return self._message

    @message.setter
    def message(self, value: str = None):

        self._message = value

    @property
    def error(self) -> str:
        """str: Property error of SensorData"""
        return self._error

    @error.setter
    def error(self, value: str = None):
        self._error = value

    @property
    def error_code(self) -> int:
        """str: Property error_code of SensorData"""
        return self._error_code

    @error_code.setter
    def error_code(self, value: int = None):
        self._error_code = value

    @property
    @validate
    def data(self) -> dict:
        """Validated data property of `SensorDataChannel`.

        Returns:
            dict: A validated dict.

        """
        sensordata = {"sensorid": self._sensor_id, "message": self._message}
        if self._error and self._error_code:
            sensordata["error"] = self._error
            sensordata["code"] = self._error_code
            return sensordata
        if len(self._channel_list) > 0:
            sensordata["channel"] = self._channel_list
        logging.debug(f"Monitoring data for sensorid {self._sensor_id}: {str(sensordata)}")
        return sensordata

    def add_channel(self, name: str, mode: str, value, **kwargs):
        """Add a channel to SensorData

        name (str): The channel name.
                mode (str): The channel mode.
                    Allowed values are "integer", "float", "counter".
                value (Any): The channel value.
                    Allowed values "integer", "float"
                **kwargs: Arbitrary keyword arguments. See utils.validators for details.

            Examples:
                >>> ping_data.add_channel(
                name="Ping Time Max",
                mode="float",
                kind="TimeResponse",
                value=float(max(ping_raw_data)),
            )

        """
        channel = self.SensorDataChannel(name=name, mode=mode, value=value, **kwargs)
        self._channel_list.append(channel.data)


class SensorProbeData(SensorData):
    """Represents SensorProbeData a special subset of SensorData"""

    def add_load_avg_channel(self, cpu_load: float, minute_avg: int):
        """Add a load average channel.

        Args:
            cpu_load (float): Load Average value.
            minute_avg (int): Minutes. Shoule be 1, 5 or 10

        """
        self.add_channel(
            name=f"Load average {str(minute_avg)} min",
            mode="float",
            kind="Custom",
            customunit="",
            value=float(cpu_load),
        )

    def add_temperature_channel(self, name: str, current_temp: float):
        """Add a temperature channel.

        Args:
            name (str): The channel name.
            current_temp: The current temperature.

        """
        self.add_channel(name=name, mode="float", kind="Temperature", value=current_temp)

    def add_disk_space_percentage_use(self, partition: str, disk_usage: list):
        """Add disk space channel.

        Args:
            partition (str): The partition name.
            disk_usage (list): List of usage values.

        """
        self.add_channel(
            name=f"Disk Space Percent {partition[1]}",
            mode="float",
            kind="Percent",
            value=disk_usage[3],
        )

    def add_disk_space_details_channel(self, name: str, partition: list, disk_usage: float):
        """Add disk space details channel,

        Args:
            name (str): The partition name.
            partition (list): List of partitions.
            disk_usage (float): The used value.

        """
        self.add_channel(
            name=f"Disk Space {name} {partition[1]}",
            mode="integer",
            kind="BytesDisk",
            value=disk_usage,
        )
